__all__ = ["configure", "export_test_set", "export_train_set", "load_test_set", "load_train_set"]
from pathlib import Path
from typing import Dict, List, Optional

from .dataset import Dataset, ImageField
from .loader import DatasetLoader

GLOBAL_LOADER = DatasetLoader()

configure = GLOBAL_LOADER.configure


def load_train_set() -> Dataset:
    """Load the training set in memory.

    MAPLES-DR training set contains labels of pathological and anatomical structures for 138 fundus images.
    The dataset is return as a :class:`Dataset` object, which is equivalent to a list of samples.
    Each sample being stored as a dictionary with the following fields:

    - ``"fundus"``: The fundus image. (MESSIDOR path must have been configured!)

    - ``"optic_cup"``: The optic cup mask.
    - ``"optic_disc"``: The optic disc mask.
    - ``"macula"``: The macula mask.
    - ``"vessels"``: The red lesions mask.
    - ``"cotton_wool_spots"``: The cotton wool spots mask.
    - ``"exudates"``: The exudates mask.
    - ``"drusens"``: The drusens mask.
    - ``"bright_uncertain"``: The mask of bright lesions with uncertain type (CWS, exudates or drusens).
    - ``"microaneurysms"``: The microaneurysms mask.
    - ``"hemorrhages"``: The hemorrhages mask.
    - ``"neovascularization"``: The neovascularization mask.
    - ``"red_uncertains"``: The mask of red lesions with uncertain type (microaneuryms or hemorrhages).
    - ``"dr"``: The Diabetic Retinopathy grade (``'R0'``, ``'R1'``, ``'R2'``, ``'R3'``, ``'R4A'``).
      See :ref:`dr-me-grades` for more information.
    - ``"me"``: The Macular Edema grade (``'M0'``, ``'M1'``, ``'M2'``).
      See :ref:`dr-me-grades` for more information.


    Returns
    -------
    Dataset
        The training set.

    Examples
    --------
    >>> import maples_dr
    >>> maples_dr.configure(messidor_path="path/to/messidor.zip")
    >>> train_set = maples_dr.load_train_set()
    >>> for sample in train_set:
    >>>     fundus = sample["fundus"]           # The fundus image
    >>>     vessels = sample["vessels"]         # The vessels mask
    >>>     dr = sample["dr"]                   # The Diabetic Retinopathy grade
    """
    return GLOBAL_LOADER.load_dataset("train")


def load_test_set() -> Dataset:
    """Load the testing set in memory.

    See :func:`load_train_set` for more details.

    Returns
    -------
    Dataset
        The testing set.
    """
    return GLOBAL_LOADER.load_dataset("test")


def export_train_set(
    path: str | Path, fields: Optional[ImageField | List[ImageField] | Dict[ImageField, str]] = None
) -> None:
    """Save the training set to a folder.



    Parameters
    ----------
    path :
        The path to the directory where
    fields :
        The field or list of fields to save. If None (by default), all fields will be saved.

        See :class:`maples_dr.dataset.ImageField` for the list of available fields.


    Examples
    --------
    >>> import maples_dr
    >>> maples_dr.configure(messidor_path="path/to/messidor.zip")
    >>> export_train_set("path/to/save", fields=["fundus", "red_lesions", "vessels"])
    """
    return GLOBAL_LOADER.load_dataset("train").export(path, fields)


def export_test_set(path: str | Path, fields: Optional[ImageField | List[ImageField] | Dict[ImageField, str]] = None):
    """Save the testing set to a folder.



    Parameters
    ----------
    path :
        The path to save the dataset to.
    fields :
        The field or list of fields to save. If None (by default), all fields will be saved.

        See :class:`maples_dr.dataset.ImageField` for the list of available fields.


    Examples
    --------
    >>> import maples_dr
    >>> maples_dr.configure(messidor_path="path/to/messidor.zip")
    >>> export_test_set("path/to/save", fields=["fundus", "red_lesions", "vessels"])
    """
    return GLOBAL_LOADER.load_dataset("test").export(path, fields)


def clear_cache():
    """Clear the cache directory."""
    return GLOBAL_LOADER.clear_cache()



from __future__ import annotations

__all__ = ["DatasetConfig", "InvalidConfigError", "ImageFormat", "Preprocessing"]
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Mapping, Optional


class ImageFormat(str, Enum):
    """String Enum of possible image formats for fundus images and biomarker masks."""

    #: Images are formatted as :class:`PIL.Image.Image`.
    PIL = "PIL"

    #: Images are formatted as numpy array.
    #: Biomarkers masks have a shape of ``(height, width)`` and
    #: fundus image have a shape of ``(height, width, 3)`` with a ``RGB`` channel order.
    RGB = "rgb"

    #: Images are formatted as numpy array.
    #: Biomarkers masks have a shape of ``(height, width)`` and
    #: fundus image have a shape of ``(height, width, 3)`` with a ``BGR`` channel order.
    BGR = "bgr"


class Preprocessing(str, Enum):
    """String Enum of possible preprocessing algorithms applied on the fundus images."""

    #: No preprocessing.
    NONE = "none"

    #: Preprocessing based on Contrast Limited Adaptive Histogram Equalization (CLAHE).
    #:
    #: See :func:`maples_dr.preprocessing.clahe_preprocessing`.
    CLAHE = "clahe"

    #: Preprocessing based on median filtering.
    #:
    #: See :func:`maples_dr.preprocessing.median_preprocessing`.
    MEDIAN = "median"


DOWNLOAD = "DOWNLOAD"


@dataclass
class DatasetConfig:
    """
    Dataclass storing the configuration of the dataset.
    """

    #: Size of the generated images.
    #:
    #: - If ``int``: the images are resized to a square of the given size;
    #: - if ``True``: the original MAPLES-DR resolution of 1500x1500 px is kept;
    #: - if ``False``: keep the original MESSIDOR resolution if MESSIDOR path is configured,
    #:   otherwise fallback to MAPLES-DR original resolution.
    #:
    #: By default: ``True`` (use MAPLES-DR 1500x1500 px resolution).
    resize: int | bool = True

    #: Python format of the generated images. See :class:`ImageFormat` for the available options.
    #:
    #: By default: :attr:`ImageFormat.PIL` is used.
    image_format: ImageFormat = ImageFormat.PIL

    #: Preprocessing algorithm applied on the fundus images.
    #: See :class:`Preprocessing` for the available options.
    #:
    #: By default: :attr:`Preprocessing.NONE` (no preprocessing) is applied.
    preprocessing: Preprocessing = Preprocessing.NONE

    _cache: str | bool = False

    def update(self, cfg: Mapping[str, Any]):
        """
        Update the configuration with the given values.

        Parameters
        ----------
        cfg:
            Configuration to update with.

        :meta private:
        """
        for k, v in cfg.items():
            if v is not None:
                setattr(self, k, v)

    @property
    def cache_path(self) -> Optional[Path]:
        """
        Return the path to the cache directory.

        Returns
        -------
            The path to the cache directory or None if the cache is disabled.
        """
        path = self._cache
        if path is False:
            return None
        return Path(path)

    def biomarkers_cache_path(
        self, biomarkers, pre_annotation: bool = False, resize: Optional[int | bool] = None
    ) -> Path:
        """Return the path to the cache directory for biomarkers masks.

        Parameters
        ----------
        biomarker: BiomarkerField | list(BiomarkerField)
            Name of the biomarker.
        pre_annotation:
            If True, the path to the cache directory for pre-annotated biomarkers masks is returned.
        resize:
            Size of the generated images. If not provided, the current size is used.

        Returns
        -------
            The path to the cache directory for biomarkers masks.

        :meta private:
        """
        from .dataset import BiomarkerField

        if isinstance(biomarkers, BiomarkerField):
            biomarkers = [biomarkers]
        folder = "+".join(b.value for b in sorted(biomarkers))
        if pre_annotation:
            folder += "_pre"
        return self.cache_path / self._resize_cache_folder_name(resize) / folder

    def fundus_cache_path(
        self, resize: Optional[int | bool] = None, preprocess: Optional[Preprocessing] = None
    ) -> Path:
        """
        Return the path to the cache directory for fundus images.

        Parameters
        ----------
        resize:
            Size of the generated images. If not provided, the current size is used.
        preprocess:
            Preprocessing algorithm applied on the fundus images. If not provided, the current preprocessing is used.

        Returns
        -------
            The path to the cache directory for fundus images.

        :meta private:
        """
        preprocess = Preprocessing(preprocess) or self.preprocessing
        preprocess_folder = "fundus" if preprocess is Preprocessing.NONE else "fundus_" + preprocess.value
        return self.cache_path / self._resize_cache_folder_name(resize) / preprocess_folder

    def _resize_cache_folder_name(self, resize: Optional[int | bool] = None) -> str:
        """
        Return the name of the cache folder for the given resize value.

        Parameters
        ----------
        resize:
            Size of the generated images. If not provided, the current size is used.

        Returns
        -------
            The name of the cache folder for the given resize value.

        :meta private:
        """
        resize = resize if resize is not None else self.resize
        if resize is False:
            return "mes"
        elif resize is True:
            return "1500"
        return str(resize)


class InvalidConfigError(Exception):
    """
    Exception raised when the configuration is invalid.
    """

    pass



from __future__ import annotations

import logging
from collections.abc import Mapping, Sequence
from enum import Enum
from pathlib import Path
from typing import Dict, Generator, Iterable, List, Optional, Tuple, TypeAlias, Union

import numpy as np
import pandas as pd
from PIL import Image
from PIL.Image import Resampling

from .config import DatasetConfig, ImageFormat, Preprocessing
from .preprocessing import fundus_roi, preprocess_fundus
from .utilities import Point, Rect, RichProgress, case_less_parse_str_enum


class BiomarkerField(str, Enum):
    """String Enum of MAPLES-DR biomarkers fields."""

    #: The optic cup mask.
    OPTIC_CUP = "opticCup"

    #: The optic disc mask.
    OPTIC_DISC = "opticDisc"

    #: The macula mask.
    MACULA = "macula"

    #: The vessels mask.
    VESSELS = "vessels"

    #: Union of all the bright lesions masks (CWS, exudates, drusens and uncertain).
    BRIGHT_LESIONS = "brightLesions"

    #: The cotton wool spots mask.
    COTTON_WOOL_SPOTS = "cottonWoolSpots"

    #: The drusens mask.
    DRUSENS = "drusens"

    #: The exudates mask.
    EXUDATES = "exudates"

    #: The mask of bright lesions with uncertain type (either CWS, exudates or drusens).
    BRIGHT_UNCERTAINS = "brightUncertains"

    #: Union of all the red lesions masks (microaneurysm, hemorrhages and uncertain).
    RED_LESIONS = "redLesions"

    #: The hemorrhages mask.
    HEMORRHAGES = "hemorrhages"

    #: The microaneurysms mask.
    MICROANEURYSMS = "microaneurysms"

    #: The neovascularization mask.
    NEOVASCULARIZATION = "neovascularization"

    #: The mask of red lesions with uncertain type (either microaneurysm or hemorrhages).
    RED_UNCERTAINS = "redUncertains"

    @classmethod
    def parse(cls, biomarker: str | BiomarkerField) -> BiomarkerField:
        """Parse a biomarker from a string.

        This method accept case insensitive biomarker names and the following aliases:
        - ``red`` for :attr:`BiomarkerField.RED_LESIONS`;
        - ``bright`` for :attr:`BiomarkerField.BRIGHT_LESIONS`;
        - ``cup`` for :attr:`BiomarkerField.OPTIC_CUP`;
        - ``opticDisk``, ``disc`` or ``disk``  for :attr:`BiomarkerField.OPTIC_DISC`;
        - ``cws`` for :attr:`BiomarkerField.COTTON_WOOL_SPOTS`;
        - ``neovessels`` for :attr:`BiomarkerField.NEOVASCULARIZATION`.

        Parameters
        ----------
        biomarker :
            The name of the biomarker.

        Returns
        -------
        BiomarkerField
            The corresponding biomarker field.

        :meta private:
        """
        return case_less_parse_str_enum(
            cls,
            biomarker,
            alias={
                "red": BiomarkerField.RED_LESIONS,
                "bright": BiomarkerField.BRIGHT_LESIONS,
                "cup": BiomarkerField.OPTIC_CUP,
                "opticdisk": BiomarkerField.OPTIC_DISC,
                "disc": BiomarkerField.OPTIC_DISC,
                "disk": BiomarkerField.OPTIC_DISC,
                "cws": BiomarkerField.COTTON_WOOL_SPOTS,
                "neovessels": BiomarkerField.NEOVASCULARIZATION,
            },
        )


AGGREGATED_BIOMARKERS = {
    BiomarkerField.BRIGHT_LESIONS: (BiomarkerField.COTTON_WOOL_SPOTS, BiomarkerField.EXUDATES, BiomarkerField.DRUSENS),
    BiomarkerField.RED_LESIONS: (BiomarkerField.MICROANEURYSMS, BiomarkerField.HEMORRHAGES),
}


class FundusField(str, Enum):
    """String Enum of MAPLES-DR fields associated with fundus images.

    .. warning::
        Path to MESSIDOR fundus images **must** be configured to use these fields!
        See :func:`maples_dr.configure` for more information.
    """

    #: The preprocessed fundus image (or the original fundus image if no preprocessing is applied).
    FUNDUS = "fundus"

    #: The raw fundus image. (If no preprocessing is applied, this is the same as :attr:`FundusField.FUNDUS`.)
    RAW_FUNDUS = "rawFundus"

    #: The mask of the fundus image.
    MASK = "mask"

    @classmethod
    def parse(cls, field: str | FundusField) -> FundusField:
        """Parse a field from a string.

        Parameters
        ----------
        field :
            The name of the field.

        Returns
        -------
        FundusField
            The corresponding field.

        :meta private:
        """
        return case_less_parse_str_enum(cls, field)


class DiagnosisField(str, Enum):
    """String Enum of MAPLES-DR diagnosis fields."""

    #: The Diabetic Retinopathy grade.
    #:
    #:  - ``R0`` : No DR.
    #:  - ``R1`` : Mild Non-Proliferative DR.
    #:  - ``R2`` : Moderate Non-Proliferative DR.
    #:  - ``R3`` : Severe Non-Proliferative DR.
    #:  - ``R4A`` : Proliferative DR.
    #:  - ``R4S`` : Treated and Stable Proliferative DR.
    #:  - ``R6`` : Insufficient quality to grade.
    DR = "dr"

    #: The Macular Edema grade.
    #:
    #:  - ``M0`` : No ME.
    #:  - ``M1`` : ME without center involvement.
    #:  - ``M2`` : ME with center involvement.
    #:  - ``M6`` : Insufficient quality to grade.
    ME = "me"

    @classmethod
    def parse(cls, field: str | DiagnosisField) -> DiagnosisField:
        """Parse a field from a string.

        Parameters
        ----------
        field :
            The name of the field.

        Returns
        -------
        DiagnosisField
            The corresponding field.

        :meta private:
        """
        return case_less_parse_str_enum(cls, field)


class BiomarkersAnnotationInfos(str, Enum):
    """Valid name for the additional information collected during the biomarkers annotations process."""

    #: The name of the retinologist who annotated the series of biomarkers.
    RETINOLOGIST = "retinologist"

    #: A comment from the retinologist.
    COMMENT = "comment"

    #: The time at which the annotation was made.
    ANNOTATION_TIME = "annotationTime"

    #: The id of the annotation. (1 is the first image to be annotated, 200 is the last one.)
    ANNOTATION_ID = "annotationID"

    @classmethod
    def parse(cls, infos: str | BiomarkersAnnotationInfos) -> BiomarkersAnnotationInfos:
        """Parse a field from a string.

        Parameters
        ----------
        infos :
            The name of the infos field.

        Returns
        -------
        BiomarkersAnnotationInfos
            The corresponding infos field.

        :meta private:
        """
        return case_less_parse_str_enum(cls, infos)


class BiomarkersAnnotationTasks(str, Enum):
    """Valid name for the different annotation tasks in MAPLES-DR."""

    #: The annotation task for bright lesions (CWS, exudates, drusens).
    BRIGHT = "brightLesions"

    #: The annotation task for red lesions (microaneurysm, hemorrhages, neovessels).
    RED = "redLesions"

    #: The annotation task for the optic disc, optic cup and the macula.
    DISC_MACULA = "discMacula"

    #: The annotation task for the vessels.
    VESSELS = "vessels"


class Pathology(str, Enum):
    """Valid name for the different pathology in MAPLES-DR."""

    #: The Diabetic Retinopathy.
    DR = "dr"

    #: The Macular Edema.
    ME = "me"


#: The name of a valid image field in MAPLES-DR.
#:
#: This type alias is a union of :class:`BiomarkerField` and :class:`FundusField`.
ImageField: TypeAlias = Union[FundusField, BiomarkerField]

#: The name of a valid field in MAPLES-DR.
#:
#: This type alias is a union of :class:`DiagnosisField`, :class:`BiomarkerField`, and :class:`FundusField`.
Field: TypeAlias = Union[DiagnosisField, ImageField]


def check_field(field: Field | str) -> Field:
    try:
        return DiagnosisField.parse(field)
    except ValueError:
        pass
    try:
        return BiomarkerField.parse(field)
    except ValueError:
        pass
    try:
        return FundusField.parse(field)
    except ValueError:
        pass

    raise ValueError(
        f"Unknown field: {field}. Must be one of:\n"
        f"  - DiagnosisField: {', '.join(_.value for _ in DiagnosisField)}\n"
        f"  - BiomarkerField: {', '.join(_.value for _ in BiomarkerField)}\n"
        f"  - FundusField: {', '.join(_.value for _ in FundusField)}\n"
    )


class Dataset(Sequence):
    """A set of samples from the MAPLES-DR dataset.

    Datasets are a utility class to access and export samples from the MAPLES-DR dataset.

    They are equivalent to a list of samples, each sample being stored as a dictionary­.
    See :obj:`Field` for the list of available fields.
    """

    def __init__(self, data: pd.DataFrame, cfg: DatasetConfig, messidor_rois: pd.DataFrame):
        """Create a new dataset. (Internal use only)

        :meta private:

        Parameters
        ----------
        data : pd.DataFrame
            The data of the dataset. It should match :attr:`Dataset.data` format.

        cfg : DatasetConfig
            The configuration of the dataset.
        messidor_rois : pd.DataFrame
            A dataframe with the same index as ``data`` containing the columns:
            - x0: the x coordinate of the top-left corner of the region of interest;
            - y0: the y coordinate of the top-left corner of the region of interest;
            - x1: the x coordinate of the bottom-right corner of the region of interest;
            - y1: the y coordinate of the bottom-right corner of the region of interest;
            - W: the width of the original MESSIDOR fundus image;
            - H: the height of the original MESSIDOR fundus image.

        """
        self._cfg: DatasetConfig = cfg

        assert isinstance(data, pd.DataFrame), "Invalid type for 'data'."
        assert all(isinstance(_, str) for _ in data.index), "Invalid index type for 'data'."
        assert all(
            col.value in data.columns for col in BiomarkerField if col not in AGGREGATED_BIOMARKERS
        ), "Missing biomarker columns in the dataset."
        assert all(col.value in data.columns for col in DiagnosisField), "Missing diagnosis columns in the dataset."
        self._data: pd.DataFrame = data

        if messidor_rois is not None:
            assert isinstance(messidor_rois, pd.DataFrame), "Invalid type for 'messidor_rois'."
            assert data.index.isin(messidor_rois.index).all(), "Invalid index for 'messidor_rois'."
            assert "x0" in messidor_rois.columns, "Missing 'x0' column in the MESSIDOR ROIs dataframe."
            assert "y0" in messidor_rois.columns, "Missing 'y0' column in the MESSIDOR ROIs dataframe."
            assert "x1" in messidor_rois.columns, "Missing 'x1' column in the MESSIDOR ROIs dataframe."
            assert "y1" in messidor_rois.columns, "Missing 'y1' column in the MESSIDOR ROIs dataframe."
            assert "W" in messidor_rois.columns, "Missing 'W' column in the MESSIDOR ROIs dataframe."
            assert "H" in messidor_rois.columns, "Missing 'H' column in the MESSIDOR ROIs dataframe."
        self._rois: pd.DataFrame = messidor_rois

    def __getitem__(self, idx: int | str | slice[int] | list[str]) -> DataSample:
        """Get a sample from the dataset.

        This method is a shortcut for :meth:`read_sample`.
        It returns the sample as a dictionary with the keys:

        - ``name`` containing the name of the image (e.g. "20051116_44816_0400_PP").
        - every :class:`Field` except ``bright_lesions`` and ``red_lesions``.

        The image are formatted following the default configuration (see :func:`maples_dr.configure`).

        Parameters
        ----------
        idx : int | str
            Index of the sample. Can be an integer or the name of the sample (e.g. "20051116_44816_0400_PP").

        Returns
        -------
        dict
            The sample as a dictionary.
        """
        if isinstance(idx, (str, int)):
            sample_infos = self.get_sample_infos(idx)
            rois = self._rois.loc[sample_infos.name]
            return DataSample(sample_infos, self._cfg, rois)
        elif isinstance(idx, slice):
            return self.subset(start=idx.start, end=idx.stop, step=idx.step)
        elif isinstance(idx, list):
            return self.subset_by_name(idx)
        else:
            raise ValueError(f"Invalid index type: {type(idx)}.")

    def __contains__(self, idx: str | int) -> bool:
        try:
            self.get_sample_infos(idx)
            return True
        except ValueError:
            return False

    def __len__(self) -> int:
        return len(self._data)

    @property
    def data(self) -> pd.DataFrame:
        """The data of the dataset.

        A dataframe containing the information of each sample. It has the following columns:
            - index: the name of the sample.
            - ``fundus``: the path to the fundus image.
            - [:class:`BiomarkerField`] (accept aggregated): the path to the biomarkers masks.
            - [:class:`BiomarkersAnnotationTasks` ``_`` :class:`BiomarkersAnnotationInfos`]: the additional annotations informations.
            - ``dr``: the consensus DR grade.
            - ``me``: the consensus ME grade.
            - ``dr_{A|B|C}``: the DR grade given by one retinologist.
            - ``me_{A|B|C}``: the ME grade given by one retinologist.
            - ``dr_{A|B|C}_comment``: comments from the retinologist when grading the DR diagnosis.
        """
        return self._data

    def keys(self) -> List[str]:
        """Get the names of the samples in the dataset.

        Returns
        -------
        List[str]
            The names of the samples.
        """
        return self._data.index.tolist()

    def subset(
        self, *arg, start: Optional[int] = None, end: Optional[int] = None, step: Optional[int] = None
    ) -> Dataset:
        """Get a subset of the dataset.

        Parameters
        ----------
        start : int
            The index of the first sample of the subset.
        end : int
            The index of the last sample of the subset.

        Returns
        -------
        Dataset
            The subset of the dataset.
        """
        if start is None and end is None:
            if step is None and len(arg) == 3:
                start, end, step = arg
            elif len(arg) == 2:
                start, end = arg
            elif len(arg) == 1:
                start = 0
                end = arg[0]
            else:
                raise ValueError("Invalid number of arguments: only expect start and end indexes.")
        elif start is None and end is not None and len(arg) == 1:
            start = arg[0]
        elif len(arg) != 0:
            raise ValueError("Invalid number of arguments: only expect start, end and step.")

        if step is None:
            step = 1

        subset_data = self._data.iloc[start:end:step]
        subset_rois = self._rois.loc[subset_data.index]
        return Dataset(subset_data, self._cfg, subset_rois)

    def subset_by_name(self, names: List[str]) -> Dataset:
        """Get a subset of the dataset by names.

        Parameters
        ----------
        names : List[str]
            The names of the samples to keep.

        Returns
        -------
        Dataset
            The subset of the dataset.
        """
        subset_data = self._data.loc[names]
        subset_rois = self._rois.loc[names]
        return Dataset(subset_data, self._cfg, subset_rois)

    def diagnosis(self, pathology: Optional[Pathology] = None):
        """Get the diagnosis of the dataset.

        Parameters
        ----------
        pathology : Pathology, optional
            The pathology to get the diagnosis for.

            If None, get the diagnosis for both pathologies.

        Returns
        -------
        pd.Series
            The diagnosis of the dataset.
        """
        retinologist = ("A", "B", "C")

        if pathology is None:
            diagnosis_fields = ["dr"] + [f"dr_{r}" for r in retinologist]
            diagnosis_fields += ["me"] + [f"me_{r}" for r in retinologist]
            diagnosis_fields += [f"dr_{r}_comment" for r in retinologist]
            return self._data[diagnosis_fields]

        pathology = Pathology(pathology)
        if pathology is Pathology.DR:
            diagnosis_fields = {"dr": "Consensus"} | {f"dr_{r}": r for r in retinologist}
        else:
            diagnosis_fields = {"me": "Consensus"} | {f"me_{r}": r for r in retinologist}
        diagnosis_fields |= {f"dr_{r}_comment": f"{r}_comment" for r in retinologist}

        return self._data[list(diagnosis_fields.keys())].rename(columns=diagnosis_fields)

    def available_fields(
        self, biomarkers=None, aggregated_biomarkers=None, diagnosis=None, fundus=None, raw_fundus=None
    ):
        any_true = (
            biomarkers is True
            or aggregated_biomarkers is True
            or diagnosis is True
            or fundus is True
            or raw_fundus is True
        )
        if not any_true:
            biomarkers = biomarkers is None
            aggregated_biomarkers = aggregated_biomarkers is None
            diagnosis = diagnosis is None
            fundus = fundus is None
            raw_fundus = raw_fundus is None

        fields = []
        if biomarkers:
            fields += [_.value for _ in BiomarkerField if _ not in AGGREGATED_BIOMARKERS]
        if aggregated_biomarkers:
            fields += list(AGGREGATED_BIOMARKERS.keys())
        if diagnosis:
            fields += [_.value for _ in DiagnosisField]
        if "fundus" in self._data.columns:
            if fundus:
                fields += [FundusField.FUNDUS.value]
            if raw_fundus:
                fields += [FundusField.RAW_FUNDUS.value]
        return fields

    def get_sample_infos(self, idx: str | int) -> pd.Series:
        """Get the information of a sample.

        Parameters
        ----------
        idx : str | int
            Index of the sample. Can be an integer or the name of the sample (i.e. "20051116_44816_0400_PP").

        Returns
        -------
        pd.Series
            The information of the sample.


        Raises
        ------
        IndexError
            If the index is out of range.

        KeyError
            If the image name is unknown.
        """
        if isinstance(idx, int):
            if idx < -len(self) or idx >= len(self):
                raise IndexError(f"Unknown sample: {idx}.")
            return self._data.iloc[idx]
        else:
            infos = self._data.loc[self._data.index == idx]
            if len(infos) == 0:
                raise KeyError(f"Unknown sample: {idx}.")
            return infos.iloc[0]

    def annotations_infos(self) -> pd.DataFrame:
        """Get the annotations information of the dataset.

        Returns
        -------
        pd.DataFrame
            The annotations information of the dataset.
        """
        tasks = [_.value for _ in BiomarkersAnnotationTasks]
        infos = [_.value for _ in BiomarkersAnnotationInfos]
        columns = pd.MultiIndex.from_product([tasks, infos], names=["Task", "Infos"])

        df = pd.DataFrame(index=self._data.index, columns=columns)
        for task in tasks:
            for info in infos:
                df[(task, info)] = self._data[task + "_" + info]
                if info == "comment":
                    df[(task, info)] = df[(task, info)].astype(str).replace("nan", "")

        return df

    def export(
        self,
        path: str | Path,
        fields: Optional[ImageField | List[ImageField] | Dict[ImageField, str]] = None,
        optimize: bool = False,
    ):
        """Export the dataset to a folder.

        Parameters
        ----------
        path :
            Path of the folder where to export the dataset.
        fields :
            The fields to be exported.

            - If None, export the whole dataset.
            - If ``fields`` is a string or list, export only the given fields.
            - If ``fields`` is a dictionary, export the fields given by the keys
            and rename their folder to their corresponding dictionary values.
        optimize :
            If True, optimize the images when exporting them.
        """
        if fields is None:
            fields = {f: f for f in self.available_fields(biomarkers=True, fundus=True, raw_fundus=True)}
        elif isinstance(fields, str):
            fields = {fields: fields}
        elif isinstance(fields, list):
            fields = {f: f for f in fields}

        if isinstance(path, str):
            path = Path(path)

        for field in fields.values():
            field_folder = path / field
            field_folder.mkdir(parents=True, exist_ok=True)

        with RichProgress.iteration("Exporting Maples-DR...", len(self), "Maples-DR exported in {t} seconds.") as p:
            for i in range(len(self)):
                sample = self.read_sample(i, list(fields.keys()), image_format="PIL")

                for field, img in sample.items():
                    if field not in fields:
                        continue

                    field_folder = path / fields[field]

                    opt = {}
                    if field not in ("fundus", "raw_fundus"):
                        opt["bits "] = 1
                        opt["optimize"] = optimize

                    img.save(field_folder / f"{sample['name']}.png", **opt)
                    p.update(1 / len(fields))


class DataSample(Mapping):
    """A sample from the MAPLES-DR dataset.

    A sample is a dictionary containing the information of a single sample from the dataset.

    """

    def __init__(
        self,
        data: pd.Series,
        cfg: DatasetConfig,
        roi: pd.Series,
    ):
        self._data: pd.Series = data
        self._cfg: DatasetConfig = cfg
        self._fundus: Optional[np.ndarray] = None
        self._roi: Rect = Rect.from_points(int(roi["y0"]), int(roi["x0"]), int(roi["y1"]), int(roi["x1"]))
        self._messidor_shape: Point = Point(int(roi["H"]), int(roi["W"]))

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, key: Field | str) -> Union[Image.Image, np.ndarray, str]:
        key = check_field(key)
        if isinstance(key, BiomarkerField):
            return self.read_biomarker(key)
        elif isinstance(key, DiagnosisField):
            return self._data[key.value]
        elif key is FundusField.FUNDUS:
            return self.read_fundus()
        elif key is FundusField.RAW_FUNDUS:
            return self.read_fundus(preprocess=False)
        elif key is FundusField.MASK:
            return self.read_roi_mask()

    def __iter__(self) -> Generator[Field]:
        if FundusField.FUNDUS.value in self._data:
            if self._cfg.preprocessing != Preprocessing.NONE:
                yield FundusField.RAW_FUNDUS
            yield [FundusField.FUNDUS]

        yield from (_ for _ in BiomarkerField if _ not in AGGREGATED_BIOMARKERS)
        yield from DiagnosisField

    def __repr__(self) -> str:
        return f"<DataSample name={self.name}>"

    @property
    def name(self) -> str:
        """The name of the sample."""
        return self._data.name

    def read_biomarker(
        self,
        biomarkers: BiomarkerField | str | Iterable[BiomarkerField | str],
        image_format: Optional[ImageFormat] = None,
        resize: Optional[int | bool] = None,
        pre_annotation: bool = False,
        no_cache: bool = False,
    ) -> any:
        """Read a biomarker from the sample.

        Parameters
        ----------
        biomarkers :
            Name of the biomarker(s) to read.

            If multiple biomarkers are given, the corresponding masks will be merged.

        image_format :
            Format of the image to return.

            If ``None`` (by default), use the format defined in the configuration.

        resize :
            Resize the image to the given size.

            - If ``resize`` is an int, crop the image to a square ROI and resize it to the shape ``(resize, resize)``;
            - If ``True``, keep the original MAPLES-DR resolution of 1500x1500 px;
            - If ``False``, use the original MESSIDOR resolution if MESSIDOR path is configured, otherwise fallback to MAPLES-DR original resolution.
            - If ``None`` (by default), use the size defined in the configuration.

        pre_annotation :
            If set to ``True``, read the pre-annotation biomarkers instead of the final ones.

            .. warning::
                Only hemorrhages, microaneurysms, exudates and vessels have pre-annotations.

        no_cache :
            If set to ``True``, the cache will not be used to read the biomarker, regardless of the configuration.

        Returns
        -------
            The biomarker mask under the format specified.
        """
        # Check arguments validity
        image_format = self._check_image_format(image_format)
        target_size, crop = self._target_size(resize)

        if isinstance(biomarkers, (str, BiomarkerField)):
            biomarkers = [biomarkers]
        biomarkers = [BiomarkerField.parse(b) for b in biomarkers]
        assert len(biomarkers), "No biomarker to read."

        # === Try to read the biomarker from the cache ===
        if not no_cache and self._cfg.cache_path is not None:
            if (
                (crop and target_size == (1500, 1500))
                and len(biomarkers) == 1
                and biomarkers[0] not in AGGREGATED_BIOMARKERS
            ):
                # If the biomarker is not aggregated and not resized, use the direct path
                key = biomarkers[0].value + ("_pre" if pre_annotation else "")
                path = self._data.get(key, None)
                if path is None or not Path(path).exists():
                    # For missing biomarkers, return an empty image
                    return Image.new("1", target_size) if image_format is ImageFormat.PIL else np.zeros(target_size)

                img = Image.open(path)
                return img if image_format is ImageFormat.PIL else np.array(img)

            # If the cache exists
            cache_path = self._cfg.biomarkers_cache_path(biomarkers, pre_annotation, resize) / f"{self.name}.png"
            if cache_path.exists():
                # Read the cached image
                img = Image.open(cache_path)
            else:
                # Otherwise read the image with the normal method and save it to the cache
                img = self.read_biomarker(
                    biomarkers=biomarkers,
                    resize=resize,
                    image_format=ImageFormat.PIL,
                    pre_annotation=pre_annotation,
                    no_cache=True,
                )
                cache_path.parent.mkdir(parents=True, exist_ok=True)
                img.save(cache_path, bits=1)

            return img if image_format is ImageFormat.PIL else np.array(img)

        # === Read the biomarker from the original files ===
        # Expand aggregated biomarkers
        for i in range(len(biomarkers) - 1, -1, -1):
            bio = biomarkers[i]
            if bio in AGGREGATED_BIOMARKERS:
                del biomarkers[i]
                biomarkers.extend(AGGREGATED_BIOMARKERS[bio])

        # Read the paths of the biomarkers
        paths = []
        for bio in biomarkers:
            key = bio.value + ("_pre" if pre_annotation else "")
            if key in self._data:
                bio_path = self._data[key]
                if bio_path is not None:
                    paths.append(bio_path)
                else:
                    if bio is BiomarkerField.MACULA:
                        logging.warning(
                            f"The macula is not segmented on image {self.name}!\n"
                            "(The corresponding fundus image is centered on the optic disc, the macula is not visible)."
                        )
                    if bio is BiomarkerField.OPTIC_CUP:
                        logging.warning(
                            f"The optic cup is not segmented on image {self.name}!\n"
                            "(The cup boundaries are too fuzzy to be annotated.)."
                        )

        # Read the biomarkers
        biomarkers_map = np.zeros(target_size, dtype=bool)
        window = biomarkers_map

        if not crop:
            # Align the image to the original MESSIDOR resolution
            window = biomarkers_map[self._roi.slice()]
            target_size = self._roi.shape

        for path in paths:
            if Path(path).exists():
                img = Image.open(path).resize(target_size.xy(), Resampling.NEAREST)
                window[np.array(img) > 0] = 1

        # Apply format conversion
        if image_format is ImageFormat.PIL:
            return Image.fromarray(biomarkers_map)
        else:
            return biomarkers_map

    def read_multiple_biomarkers(
        self,
        biomarkers: Mapping[int, BiomarkerField | str | List[BiomarkerField | str]],
        image_format: Optional[ImageFormat] = None,
        pre_annotation: bool = False,
        resize: Optional[int | bool] = None,
    ):
        """
        Read multiple biomarkers at once, assigning a class to a biomarker or a group of them.

        Parameters
        ----------
        biomarkers :
            Name of the biomarker(s) to read.

        image_format :
            Format of the image to return.

            If None (by default), use the format defined in the configuration.

        pre_annotation :
            If set to ``True``, read the pre-annotation biomarkers instead of the final ones.

            .. warning::
                Only hemorrhages, microaneurysms, exudates and vessels have pre-annotations.

        """
        image_format = self._check_image_format(image_format)
        target_size, _ = self._target_size(resize)

        biomarkers_map = np.zeros(target_size, dtype=np.uint8)
        for i, bio in biomarkers.items():
            biomarkers_map[self.read_biomarker(bio, image_format="bgr", pre_annotation=pre_annotation)] = i

        # Apply format conversion
        if image_format is ImageFormat.PIL:
            return Image.fromarray(biomarkers_map)
        else:
            return biomarkers_map

    def read_fundus(
        self,
        preprocess: Optional[Preprocessing | str | bool] = None,
        image_format: Optional[ImageFormat] = None,
        resize: Optional[int | bool] = None,
        no_cache: bool = False,
    ) -> Union[Image.Image, np.ndarray]:
        """Read the fundus image of the sample.

        Parameters
        ----------
        preprocess :
            Preprocessing to apply to the image.

            - If a :class:`maples_dr.config.Preprocessing` (or an equivalent string), the image is preprocessed with the given preprocessing;
            - if ``False``, the image is not preprocessed.
            - if ``None`` (by default) or ``True``, use the preprocessing defined in the configuration.

        image_format :
            Format of the image to return.

            If ``None`` (by default), use the format defined in the configuration.

        resize :
            Resize the image to the given size.

            - If ``resize`` is an int, crop the image to a square ROI and resize it to the shape ``(resize, resize)``;
            - If ``True``, use the original MAPLES-DR resolution of 1500x1500 px;
            - If ``False``, keep the original MESSIDOR resolution.
            - If ``None`` (by default), use the size defined in the configuration.

        no_cache :
            If set to ``True``, the cache will not be used to read the fundus image, regardless of the configuration.

        Returns
        -------
            The fundus image under the format specified.
        """

        assert "fundus" in self._data, "Impossible to read fundus images, path to MESSIDOR dataset is unknown."

        # Check arguments
        preprocess = self._check_preprocessing(preprocess)
        image_format = self._check_image_format(image_format)
        target_size, crop = self._target_size(resize)

        # Check if the result is cached
        cache_path = self._cfg.cache_path
        if not no_cache and cache_path is not None:
            if not crop and preprocess is Preprocessing.NONE:
                img = Image.open(self._data[FundusField.FUNDUS.value])
            else:
                cache_path = self._cfg.fundus_cache_path(resize, preprocess) / f"{self.name}.jpg"

                if not cache_path.exists():
                    img = self.read_fundus(
                        preprocess=preprocess, resize=resize, image_format=ImageFormat.PIL, no_cache=True
                    )
                    cache_path.parent.mkdir(parents=True, exist_ok=True)
                    img.save(cache_path, quality=98, subsampling=0)

                # Even if we just created the cache, we reload it to ensure consistency despite compression losses
                img = Image.open(cache_path)

            return self._apply_image_format(img, ImageFormat.PIL, image_format)

        # Read the image
        if self._fundus is None:
            path = self._data[FundusField.FUNDUS.value]
            fundus = Image.open(path)
            fundus_format = ImageFormat.PIL

            # Crop the image to the proper region of interest
            if crop and fundus.height != fundus.width:
                fundus = fundus.crop(self._roi.box())
        else:
            fundus = self._fundus
            fundus_format = ImageFormat.BGR

        # Resize the image
        fundus_shape = fundus.size if fundus_format is ImageFormat.PIL else fundus.shape[:2]
        if fundus_shape != target_size:
            resampling_method = Resampling.LANCZOS if target_size.y < 1000 else Resampling.BILINEAR
            if fundus_format is not ImageFormat.PIL:
                fundus = self._apply_image_format(fundus, fundus_format, ImageFormat.PIL)
                fundus_format = ImageFormat.PIL
            fundus = fundus.resize(target_size.xy(), resampling_method)

        # Preprocess the image
        if preprocess is not Preprocessing.NONE:
            # If required, cast the image to a numpy array
            fundus = self._apply_image_format(fundus, fundus_format, ImageFormat.BGR)
            fundus_format = ImageFormat.BGR
            self._fundus = fundus

            fundus = preprocess_fundus(fundus, preprocess)

        return self._apply_image_format(fundus, fundus_format, image_format)

    def read_roi_mask(self, image_format: Optional[ImageFormat] = None) -> np.ndarray | Image.Image:
        """Read the region of interest of the fundus image.

        Parameters
        ----------

        image_format :
            Format of the image to return. If None, use the format defined in the configuration.

        Returns
        -------
        np.ndarray
            The region of interest of the fundus image.
        """
        fundus = self.read_fundus(preprocess=False, image_format=ImageFormat.BGR)
        mask = fundus_roi(fundus)

        return self._apply_image_format(mask, ImageFormat.BGR, image_format)

    def _apply_image_format(
        self, image: Image.Image | np.ndarray, input_format: ImageFormat, target_format: str | ImageFormat | None
    ) -> Image.Image | np.ndarray:
        target_format = self._check_image_format(target_format)
        input_format = ImageFormat(input_format)

        if input_format is ImageFormat.PIL:
            if target_format is ImageFormat.PIL:
                return image
            else:  # Target format is numpy array
                image = np.array(image)
                if len(image.shape) == 2 or target_format is ImageFormat.RGB:
                    return image
                else:
                    return image[..., ::-1]
        else:  # Input format is numpy array
            if target_format is ImageFormat.PIL:
                return Image.fromarray(image)
            else:  # Target format is numpy array
                same_channel_order = len(image.shape) == 2 or target_format is input_format
                return image if same_channel_order else image[..., ::-1]

    def _check_image_format(self, image_format: Optional[ImageFormat | str] = None) -> ImageFormat:
        if image_format is None:
            return ImageFormat(self._cfg.image_format)
        return ImageFormat(image_format)

    def _check_preprocessing(self, preprocess: str | Preprocessing | bool | None) -> Preprocessing:
        if preprocess is None or preprocess is True:
            return Preprocessing(self._cfg.preprocessing)
        if preprocess is False:
            return Preprocessing.NONE
        return Preprocessing(preprocess)

    def _target_size(self, size: Optional[int | bool] = None) -> Tuple[Point, bool]:
        if size is None:
            size = self._cfg.resize

        if size is False:
            return self._messidor_shape, False
        elif size is True:
            return Point(1500, 1500), True
        else:
            return Point(size, size), True



from typing import Optional

import numpy as np

from .config import Preprocessing

try:
    import cv2
except ImportError:
    cv2 = None

try:
    from skimage import measure as skmeasure
    from skimage import morphology as skmorph
except ImportError:
    skmeasure = None
    skmorph = None


__all__ = ["preprocess_fundus", "clahe_preprocessing", "median_preprocessing", "fundus_roi"]


def ensure_imports(cv2_needed=False, skimage_needed=False):
    if cv2_needed and skimage_needed and (cv2 is None and skmeasure is None):
        raise ImportError(
            "OpenCV and Scikit-Image are required for this function.\n"
            "Please install them using 'pip install opencv-python-headless scikit-image'."
        )
    if cv2_needed and not cv2:
        raise ImportError(
            "OpenCV is required for this function." "Please install it using 'pip install opencv-python-headless'."
        )
    if skimage_needed and not skmeasure:
        raise ImportError(
            "Scikit-Image is required for this function." "Please install it using 'pip install scikit-image'."
        )


def preprocess_fundus(fundus: np.ndarray, preprocessing: Preprocessing | str) -> np.ndarray:
    """Preprocess a fundus image.

    Parameters
    ----------
    fundus:
        The fundus image to preprocess.
    preprocessing:
        The preprocessing algorithm to apply.
        See :class:`maples_dr.config.Preprocessing` for the available options.

    Returns
    -------
        The preprocessed fundus image.

    """
    preprocessing = Preprocessing(preprocessing)

    if preprocessing is Preprocessing.NONE:
        return fundus
    elif preprocessing is Preprocessing.CLAHE:
        return clahe_preprocessing(fundus)
    elif preprocessing is Preprocessing.MEDIAN:
        return median_preprocessing(fundus)


def clahe_preprocessing(fundus: np.ndarray, mask: Optional[np.ndarray] = None) -> np.ndarray:
    """Preprocessing based on Contrast Limited Adaptive Histogram Equalization (CLAHE).

    This algorithm was used to annotate MAPLES-DR anatomical and pathological structures.

    Parameters
    ----------
    fundus:
        The fundus image as a BGR numpy array (height, width, 3).

    Returns
    -------
        The preprocessed fundus image.

    """
    # Initial checks
    assert isinstance(fundus, np.ndarray), f"Fundus image must be a numpy array instead of {type(fundus)}."
    assert fundus.dtype == np.uint8, f"Fundus image must be a 8-bit unsigned integer array instead of {fundus.dtype}."
    assert (
        len(fundus.shape) == 3 and fundus.shape[2] == 3
    ), f"Fundus image must be a 3-channel RGB image with shape (height, width, 3) instead of {fundus.shape}."
    mask = fundus_roi(fundus) if mask is None else mask
    assert mask.shape == fundus.shape[:2], "Mask must have the same shape as the fundus image."

    # CV2 is required for this preprocessing
    ensure_imports(cv2_needed=True)

    # Preprocessing
    mean_b = np.median(fundus[..., 0][mask])
    mean_g = np.median(fundus[..., 1][mask])
    mean_r = np.median(fundus[..., 2][mask])
    mean_channels = np.asarray([mean_b, mean_g, mean_r])

    mask = np.expand_dims(mask, 2)

    bgr = fundus.astype(np.float32)
    bgr += (mean_channels - cv2.medianBlur(fundus, 151)) * mask
    bgr = bgr.clip(0, 255).astype(np.uint8)

    lab = cv2.cvtColor(bgr, cv2.COLOR_BGR2LAB)
    lab_layers = list(cv2.split(lab))
    clahe = cv2.createCLAHE(clipLimit=2.0, tileGridSize=(8, 8))
    lab_layers[0] = clahe.apply(lab_layers[0])
    lab = cv2.merge(lab_layers)
    bgr = cv2.cvtColor(lab, cv2.COLOR_LAB2BGR)

    return bgr * mask


def median_preprocessing(fundus: np.ndarray) -> np.ndarray:
    """Preprocessing based on median filtering.

    This algorithm is often used as a preprocessing step for automatic vessel segmentation.

    Parameters
    ----------
    fundus:
        The fundus image as a BGR numpy array (height, width, 3).

    Returns
    -------
        The preprocessed fundus image.

    """
    assert isinstance(fundus, np.ndarray), f"Fundus image must be a numpy array instead of {type(fundus)}."
    assert fundus.dtype == np.uint8, f"Fundus image must be a 8-bit unsigned integer array instead of {fundus.dtype}."
    assert (
        len(fundus.shape) == 3 and fundus.shape[2] == 3
    ), "Fundus image must be a 3-channel RGB image with shape (height, width, 3) instead of {fundus.shape}."

    # CV2 is required for this preprocessing
    ensure_imports(cv2_needed=True)

    k = np.max(fundus.shape) // 20 * 2 + 1
    bg = cv2.medianBlur(fundus, k)
    return cv2.addWeighted(fundus, 4, bg, -4, 128)


def fundus_roi(
    fundus: np.ndarray, blur_radius=5, morphological_clean=False, smoothing_radius=0, final_erosion=4
) -> np.ndarray:
    """Compute the region of interest (ROI) of a fundus image.

    Parameters:
    -----------
    fundus:
        The fundus image.

    blur_radius:
        The radius of the median blur filter.

        By default: 5.

    morphological_clean:
        Whether to perform morphological cleaning. (small objects removal and filling of the holes not on the border)

        By default: False.

    smoothing_radius:
        The radius of the Gaussian blur filter.

        By default: 0.

    final_erosion:
        The radius of the disk used for the final erosion.

        By default: 4.

    Returns:
        The ROI mask.

    """
    ensure_imports(skimage_needed=True, skimage_needed=morphological_clean or final_erosion > 0)

    fundus = cv2.medianBlur(fundus[..., 1], blur_radius * 2 - 1)
    mask = fundus > 10

    if morphological_clean:
        # Remove small objects
        mask = skmorph.remove_small_objects(mask, 5000)

        # Remove holes that are not on the border
        MASK_BORDER = np.zeros_like(mask)
        MASK_BORDER[0, :] = 1
        MASK_BORDER[-1, :] = 1
        MASK_BORDER[:, 0] = 1
        MASK_BORDER[:, -1] = 1
        labelled_holes = skmeasure.label(mask == 0)
        for i in range(1, labelled_holes.max() + 1):
            hole_mask = labelled_holes == i
            if not np.any(MASK_BORDER & hole_mask):
                mask[hole_mask] = 1

    if smoothing_radius > 0:
        mask = (
            cv2.GaussianBlur(
                mask.astype(np.uint8) * 255,
                (smoothing_radius * 6 + 1, smoothing_radius * 6 + 1),
                smoothing_radius,
                borderType=cv2.BORDER_CONSTANT,
            )
            > 125
        )

    if final_erosion > 0:
        skmorph.binary_erosion(mask, skmorph.disk(final_erosion), out=mask)

    return mask



import logging
from time import time

import maples_dr
import numpy as np
from maples_dr.dataset import (
    BiomarkerField,
)
from pytest import fixture

LOGGER = logging.getLogger(__name__)


@fixture
def train_set():
    maples_dr.configure(
        maples_dr_path="examples/PATH/TO/MAPLES-DR/AdditionalData.zip",
        messidor_path="examples/PATH/TO/MESSIDOR/",
    )
    return maples_dr.load_train_set()


def test_cache_efficiency_fundus(train_set, tmp_path):
    N_SAMPLES = 50

    # Measure Timing without cache
    maples_dr.configure(cache=False, image_format="bgr")

    tic = time()
    fundus_nocache = [sample.read_fundus() for sample in train_set[:N_SAMPLES]]
    dt_nocache = time() - tic

    # Measure Timing with cache
    maples_dr.configure(cache=tmp_path)

    # A first time to create the cache
    tic = time()
    fundus_cache0 = [sample.read_fundus() for sample in train_set[:N_SAMPLES]]
    dt_cache0 = time() - tic

    # A second time to use the cache
    tic = time()
    fundus_cache1 = [sample.read_fundus() for sample in train_set[:N_SAMPLES]]
    dt_cache1 = time() - tic

    # A third time to check consistency
    tic = time()
    fundus_cache2 = [sample.read_fundus() for sample in train_set[:N_SAMPLES]]
    dt_cache2 = time() - tic

    # Check that the cache data is consistent
    def tolerable_diff(a, b):
        return np.sum(np.abs(a.astype(float) - b) > 3) < a.size / 100

    assert all(
        tolerable_diff(a, b) for a, b in zip(fundus_cache1, fundus_cache0, strict=True)
    ), "Cache data inconsistent between cache creation and usage."
    assert all(
        tolerable_diff(a, b) for a, b in zip(fundus_cache1, fundus_cache2, strict=True)
    ), "Cache data inconsistent between cache usage."
    assert all(
        tolerable_diff(a, b) for a, b in zip(fundus_cache1, fundus_nocache, strict=True)
    ), "Cache data inconsistent with standard read."

    # Check that the cache is time effective
    assert max(dt_cache1, dt_cache2) < dt_cache0
    assert max(dt_cache1, dt_cache2) < dt_nocache

    # Print timing information
    LOGGER.info(f"Without cache: {dt_nocache/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"With cache: {dt_cache1/N_SAMPLES*1000:.1f}ms , {dt_cache2/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache creation: {dt_cache0/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache efficiency: {dt_nocache / max(dt_cache1, dt_cache2):.2f}")


def test_cache_efficiency_vessels(train_set, tmp_path):
    N_SAMPLES = 50

    # === COMPARE TIMING WHEN NO PROCESSING IS APPLIED ===
    # Measure Timing without cache
    maples_dr.configure(cache=False, image_format="bgr")

    tic = time()
    vessels_nocache = [sample.read_biomarker(BiomarkerField.VESSELS) for sample in train_set[:N_SAMPLES]]
    dt_nocache = time() - tic

    # Measure Timing with cache
    maples_dr.configure(cache=tmp_path)
    maples_dr.clear_cache()
    tic = time()
    vessels_cache0 = [sample.read_biomarker(BiomarkerField.VESSELS) for sample in train_set[:N_SAMPLES]]
    dt_cache0 = time() - tic

    # Check that the cache data is consistent
    assert all(
        np.all(a == b) for a, b in zip(vessels_cache0, vessels_nocache, strict=True)
    ), "Cache data inconsistent with standard read."

    # Check that the cache is time effective
    assert (
        (dt_cache0 - dt_nocache) / dt_nocache < 1e-2
    ), f"Cache read is slower than no cache read (no cache: {dt_nocache:.2f}s, cache: {dt_cache0:.2f}s)"

    # Print timing information
    LOGGER.info("=== NO PROCESSING ===")
    LOGGER.info(f"Without cache: {dt_nocache/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"With cache: {dt_cache0/N_SAMPLES*1000:.1f}ms")

    # === COMPARE TIMING WHEN PROCESSING IS APPLIED ===
    maples_dr.configure(resize=512)

    # Measure Timing without cache
    maples_dr.configure(cache=False, image_format="bgr")
    tic = time()
    vessels_nocache = [sample.read_biomarker(BiomarkerField.VESSELS) for sample in train_set[:N_SAMPLES]]
    dt_nocache = time() - tic

    # Measure Timing with cache
    maples_dr.configure(cache=tmp_path)

    # A first time to create the cache
    tic = time()
    vessels_cache0 = [sample.read_biomarker(BiomarkerField.VESSELS) for sample in train_set[:N_SAMPLES]]
    dt_cache0 = time() - tic

    # A second time to use the cache
    tic = time()
    vessels_cache1 = [sample.read_biomarker(BiomarkerField.VESSELS) for sample in train_set[:N_SAMPLES]]
    dt_cache1 = time() - tic

    # A third time to check consistency
    tic = time()
    vessels_cache2 = [sample.read_biomarker(BiomarkerField.VESSELS) for sample in train_set[:N_SAMPLES]]
    dt_cache2 = time() - tic

    # Check that the cache data is consistent
    assert all(
        np.all(a == b) for a, b in zip(vessels_cache1, vessels_cache0, strict=True)
    ), "Cache data inconsistent between cache creation and usage."
    assert all(
        np.all(a == b) for a, b in zip(vessels_cache1, vessels_cache2, strict=True)
    ), "Cache data inconsistent between cache usage."
    assert all(
        np.all(a == b) for a, b in zip(vessels_cache1, vessels_nocache, strict=True)
    ), "Cache data inconsistent with standard read."

    # Check that the cache is time effective
    assert (
        max(dt_cache1, dt_cache2) < dt_cache0
    ), f"Cache read is slower than cache creation (create: {dt_cache0:.2f}s, read: {dt_cache1:.2f}s, {dt_cache2:.2f}s)"
    assert (
        max(dt_cache1, dt_cache2) < dt_nocache
    ), f"Cache read is slow (cache: {dt_cache1:.2f}s, {dt_cache2:.2f}s, no cache: {dt_nocache:.2f}s)"

    # Print timing information
    LOGGER.info("=== WITH PROCESSING ===")
    LOGGER.info(f"Without cache: {dt_nocache/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"With cache: {dt_cache1/N_SAMPLES*1000:.1f}ms , {dt_cache2/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache creation: {dt_cache0/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache efficiency: {dt_nocache / max(dt_cache1, dt_cache2):.2f}")


def test_cache_efficiency_red(train_set, tmp_path):
    N_SAMPLES = 50

    # === COMPARE TIMING WHEN NO PROCESSING IS APPLIED ===
    # Measure Timing without cache
    maples_dr.configure(cache=False, image_format="bgr")

    tic = time()
    red_nocache = [sample.read_biomarker(BiomarkerField.RED_LESIONS) for sample in train_set[:N_SAMPLES]]
    dt_nocache = time() - tic

    # Measure Timing with cache
    maples_dr.configure(cache=tmp_path)
    maples_dr.clear_cache()

    # A first time to create the cache
    tic = time()
    red_cache0 = [sample.read_biomarker(BiomarkerField.RED_LESIONS) for sample in train_set[:N_SAMPLES]]
    dt_cache0 = time() - tic

    # A second time to use the cache
    tic = time()
    red_cache1 = [sample.read_biomarker(BiomarkerField.RED_LESIONS) for sample in train_set[:N_SAMPLES]]
    dt_cache1 = time() - tic

    # Check that the cache data is consistent
    assert all(
        np.all(a == b) for a, b in zip(red_cache1, red_cache0, strict=True)
    ), "Cache data inconsistent between cache creation and usage."
    assert all(
        np.all(a == b) for a, b in zip(red_cache0, red_nocache, strict=True)
    ), "Cache data inconsistent with standard read."

    # Check that the cache is time effective
    assert (
        dt_cache1 < dt_cache0
    ), f"Cache read is slower than cache creation (create: {dt_cache0:.2f}s, read: {dt_cache1:.2f}s)"
    assert (
        (dt_cache1 - dt_nocache) / dt_nocache < 1e-2
    ), f"Cache read is slower than no cache read (no cache: {dt_nocache:.2f}s, cache: {dt_cache0:.2f}s)"

    # Print timing information
    LOGGER.info("=== NO PROCESSING ===")
    LOGGER.info(f"Without cache: {dt_nocache/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"With cache: {dt_cache1/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache creation: {dt_cache0/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache efficiency: {dt_nocache / dt_cache1:.2f}")

    # === COMPARE TIMING WHEN PROCESSING IS APPLIED ===
    maples_dr.configure(resize=512)

    # Measure Timing without cache
    maples_dr.configure(cache=False, image_format="bgr")
    tic = time()
    red_nocache = [sample.read_biomarker(BiomarkerField.RED_LESIONS) for sample in train_set[:N_SAMPLES]]
    dt_nocache = time() - tic

    # Measure Timing with cache
    maples_dr.configure(cache=tmp_path)

    # A first time to create the cache
    tic = time()
    red_cache0 = [sample.read_biomarker(BiomarkerField.RED_LESIONS) for sample in train_set[:N_SAMPLES]]
    dt_cache0 = time() - tic

    # A second time to use the cache
    tic = time()
    red_cache1 = [sample.read_biomarker(BiomarkerField.RED_LESIONS) for sample in train_set[:N_SAMPLES]]
    dt_cache1 = time() - tic

    # A third time to check consistency
    tic = time()
    red_cache2 = [sample.read_biomarker(BiomarkerField.RED_LESIONS) for sample in train_set[:N_SAMPLES]]
    dt_cache2 = time() - tic

    # Check that the cache data is consistent
    assert all(
        np.all(a == b) for a, b in zip(red_cache1, red_cache0, strict=True)
    ), "Cache data inconsistent between cache creation and usage."
    assert all(
        np.all(a == b) for a, b in zip(red_cache1, red_cache2, strict=True)
    ), "Cache data inconsistent between cache usage."
    assert all(
        np.all(a == b) for a, b in zip(red_cache1, red_nocache, strict=True)
    ), "Cache data inconsistent with standard read."

    # Check that the cache is time effective
    assert (
        max(dt_cache1, dt_cache2) < dt_cache0
    ), f"Cache read is slower than cache creation (create: {dt_cache0:.2f}s, read: {dt_cache1:.2f}s, {dt_cache2:.2f}s)"
    assert (
        max(dt_cache1, dt_cache2) < dt_nocache
    ), f"Cache read is slow (cache: {dt_cache1:.2f}s, {dt_cache2:.2f}s, no cache: {dt_nocache:.2f}s)"

    # Print timing information
    LOGGER.info("=== WITH PROCESSING ===")
    LOGGER.info(f"Without cache: {dt_nocache/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"With cache: {dt_cache1/N_SAMPLES*1000:.1f}ms , {dt_cache2/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache creation: {dt_cache0/N_SAMPLES*1000:.1f}ms")
    LOGGER.info(f"Cache efficiency: {dt_nocache / max(dt_cache1, dt_cache2):.2f}")



import logging

import maples_dr
from maples_dr.dataset import BiomarkerField as Bio
from maples_dr.dataset import DiagnosisField
from pytest import fixture

LOGGER = logging.getLogger(__name__)


@fixture
def local_dataset():
    local_loader = maples_dr.loader.DatasetLoader()
    local_loader.configure(
        maples_dr_path="examples/PATH/TO/MAPLES-DR/AdditionalData.zip",
    )
    return local_loader.load_dataset("all_with_duplicates")


def test_download(local_dataset):
    downloaded_dataset = maples_dr.quick_api.GLOBAL_LOADER.load_dataset("all_with_duplicates")

    # === Check data content ===
    # Biomarkers
    for local_sample, downloaded_sample in zip(local_dataset, downloaded_dataset, strict=True):
        for biomarker in Bio:
            assert (
                local_sample[biomarker] == downloaded_sample[biomarker]
            ), f"Inconsistent {biomarker} for sample {local_sample.name}."

        for field in DiagnosisField:
            assert local_sample[field] == downloaded_sample[field], (
                f"Inconsistent {field} for sample {local_sample.name}."
                f" (local: {local_sample[field]}, downloaded: {downloaded_sample[field]})",
            )



import logging

import maples_dr
import numpy as np
from maples_dr.dataset import (
    AGGREGATED_BIOMARKERS,
    BiomarkersAnnotationInfos,
    BiomarkersAnnotationTasks,
    DiagnosisField,
)
from maples_dr.dataset import BiomarkerField as Bio
from pytest import fixture

LOGGER = logging.getLogger(__name__)


@fixture
def train_set():
    maples_dr.configure(
        maples_dr_path="examples/PATH/TO/MAPLES-DR/AdditionalData.zip",
    )
    return maples_dr.load_train_set()


@fixture
def test_set():
    maples_dr.configure(
        maples_dr_path="examples/PATH/TO/MAPLES-DR/AdditionalData.zip",
    )
    return maples_dr.load_test_set()


@fixture
def all_set():
    maples_dr.configure(
        maples_dr_path="examples/PATH/TO/MAPLES-DR/AdditionalData.zip",
    )
    return maples_dr.quick_api.GLOBAL_LOADER.load_dataset("all_with_duplicates")


def test_dataset_properties(train_set, test_set, all_set):
    assert len(train_set) == 138
    assert len(test_set) == 60
    assert len(all_set) == 200

    # === Check data content ===
    # Biomarkers
    for biomarker in Bio:
        if biomarker not in AGGREGATED_BIOMARKERS:
            assert biomarker.value in all_set.data
    for biomarker in [
        Bio.EXUDATES,
        Bio.HEMORRHAGES,
        Bio.MICROANEURYSMS,
        Bio.VESSELS,
    ]:
        assert f"{biomarker.value}_pre" in all_set.data

    # Biomarkers tasks
    for task in BiomarkersAnnotationTasks:
        for infos in BiomarkersAnnotationInfos:
            assert f"{task.value}_{infos.value}" in all_set.data

    # Diagnosis
    for r in "ABC":
        for pathology in ("dr", "me"):
            grades = ["R0", "R1", "R2", "R3", "R4A", "R4S", "R6"] if pathology == "dr" else ["M0", "M1", "M2", "M6"]

            assert pathology in all_set.data
            invalid_values = set(all_set.data[pathology].unique()) - set(grades)
            assert not invalid_values, f"Invalid values for {pathology}: {invalid_values}"

            assert f"{pathology}_{r}" in all_set.data
            invalid_values = set(all_set.data[f"{pathology}_{r}"].unique()) - set(grades)
            assert not invalid_values, f"Invalid values for {pathology}_{r}: {invalid_values}"

        assert f"dr_{r}_comment" in all_set.data


def test_dataset_numerical_indexing(all_set):
    names0 = all_set.keys()

    assert all_set[0].name == names0[0]
    assert all_set[-1].name == names0[-1]
    assert all_set[:5].keys() == names0[:5]
    assert all_set[5:10].keys() == names0[5:10]
    assert all_set[-5:].keys() == names0[-5:]


def test_dataset_samples(all_set):
    sample = all_set[0]
    sample._cfg.image_format = "bgr"

    assert sample[Bio.VESSELS].sum() > 200

    # Test equivalence of __getitem__ and actual read methods for biomarkers
    for biomarker in Bio:
        assert np.all(sample[biomarker] == sample.read_biomarker(biomarker))
        assert sample.read_biomarker(biomarker, pre_annotation=True) is not None

        # Test aggregated biomarkers
        if biomarker in AGGREGATED_BIOMARKERS:
            no_diff = sample[biomarker] == np.any(
                [sample[bio] for bio in AGGREGATED_BIOMARKERS[biomarker]],
                axis=0,
            )
            assert np.all(no_diff)

    # Test field name tolerance
    assert np.all(sample["CottonWoolSpots"] == sample["cotton_wool_spots"])
    assert np.all(sample["red"] == sample["RedLesions"])
    assert np.all(sample["bright"] == sample["bright_lesions"])
    assert np.all(sample["CUP"] == sample["OPTIC_CUP"])
    assert np.all(sample["Disk"] == sample["opticDisk"])

    # Test existence of diagnosis fields
    assert sample[DiagnosisField.DR] in ("R0", "R1", "R2", "R3", "R4A")
    assert sample[DiagnosisField.ME] in ("M0", "M1", "M2")

    # Test existence of diagnosis infos fields
    for r in "ABC":
        for pathology in ("dr", "me"):
            assert pathology in all_set.data
            assert f"{pathology}_{r}" in all_set.data
        assert f"dr_{r}_comment" in all_set.data


def test_anatomical_structures_on_all_samples(all_set, caplog):
    maples_dr.configure(image_format="bgr")
    missing_vessels = set()
    missing_disc = set()
    missing_cup = set()
    missing_macula = set()
    for sample in all_set:
        if sample[Bio.VESSELS].sum() == 0:
            missing_vessels.add(sample.name)
        if sample[Bio.OPTIC_DISC].sum() == 0:
            missing_disc.add(sample.name)

        caplog.clear()
        if sample[Bio.MACULA].sum() == 0:
            if sample.name not in maples_dr.quick_api.GLOBAL_LOADER.dataset_record.get("no_macula", []):
                missing_macula.add(sample.name)
            else:
                assert (
                    f"The macula is not segmented on image {sample.name}!" in caplog.text
                ), "Missing macula should be logged."
        if sample[Bio.OPTIC_CUP].sum() == 0:
            if sample.name not in maples_dr.quick_api.GLOBAL_LOADER.dataset_record.get("no_cup", []):
                missing_cup.add(sample.name)
            else:
                assert (
                    f"The optic cup is not segmented on image {sample.name}!" in caplog.text
                ), "Missing cup should be logged."

    assert not missing_vessels, f"Missing vessels for images {missing_vessels}."
    assert not missing_disc, f"Missing disc for images {missing_disc}."
    assert not missing_macula, f"Missing macula for images {missing_macula}."
    assert not missing_cup, f"Missing cup for images {missing_cup}."


def test_lesions_on_all_pathological_samples(all_set):
    maples_dr.configure(image_format="bgr")

    missing_lesions = set()
    for sample in all_set:
        if sample["dr"] not in ("R0", "R6") or sample["me"] not in ("M0", "M6"):
            if sample.read_biomarker([Bio.RED_LESIONS, Bio.BRIGHT_LESIONS]).sum() == 0:
                missing_lesions.add(sample.name)

    assert not missing_lesions, f"Missing lesions for {len(missing_lesions)} images {missing_lesions}."



__all__ = ["configure", "Dataset", "export_test_set", "export_train_set", "load_test_set", "load_train_set"]

import os

if os.environ.get("MAPLES-DR_SILENT_IMPORT", None) is None:
    from rich.console import Console

    console = Console()
    console.print(
        "[b]Thanks for using MAPLES-DR![/b]\n"
        "  When using this dataset in academic works,\n"
        "  please cite: [u cyan link=https://arxiv.org/abs/2402.04258]https://arxiv.org/abs/2402.04258[/u cyan link]"
    )

from .dataset import DataSample, Dataset
from .quick_api import clear_cache, configure, export_test_set, export_train_set, load_test_set, load_train_set



__all__ = ["DatasetLoader", "NotConfiguredError", "UNSET"]
import shutil
from enum import Enum
from functools import partial
from itertools import chain
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import Dict, Optional, Tuple
from urllib.request import urlopen
from zipfile import ZipFile

import pandas as pd
import yaml

from .config import (
    DOWNLOAD,
    DatasetConfig,
    ImageFormat,
    InvalidConfigError,
    Preprocessing,
)
from .dataset import (
    BiomarkerField,
    BiomarkersAnnotationInfos,
    BiomarkersAnnotationTasks,
    Dataset,
    FundusField,
)
from .utilities import RichProgress, xdg_data_home

#   === CONSTANTS ===
# Figshare public urls of the MAPLES DR dataset
MAPLES_DR_ADDITIONAL_URL = "https://figshare.com/ndownloader/files/45795384"

#: Unset constant
UNSET = "UNSET"


class DatasetSubset(str, Enum):
    """Enumeration of the possible subsets of the MAPLES-DR dataset."""

    #: The training set (138 images)
    TRAIN = "train"

    #: The testing set (60 images)
    TEST = "test"

    #: All the images of the training and test set (198 images)
    ALL = "all"

    #: The two duplicated images of the training set
    DUPLICATES = "duplicates"

    #: All the images of the training and test set, including duplicates (200 images)
    ALL_WITH_DUPLICATES = "all_with_duplicates"


class NotConfiguredError(Exception):
    """
    Exception raised when the dataset loader is not configured.
    """

    def __init__(
        self, message: str = "MAPLES-DR dataset is not configured yet.", *args
    ):
        super().__init__(message, *args)


class DatasetLoader:
    """
    Loader for MAPLES-DR dataset.
    """

    def __init__(self):
        self._datasets_config = DatasetConfig(
            resize=1500,
            image_format="PIL",
            preprocessing="none",
        )
        self.dataset_record: Optional[dict] = None
        self._maples_dr_path: Optional[Path] = None
        self._diagnosis: Optional[pd.DataFrame] = None
        self._annotations_infos: Optional[Dict[str, pd.DataFrame]] = None
        self._is_maples_dr_folder_temporary: bool = False

        self._messidor_paths: Optional[dict[str, str]] = None
        self._messidor_ROIs: Optional[pd.DataFrame] = None

    def __del__(self):
        if self._is_maples_dr_folder_temporary:
            rmtree(self._maples_dr_path, ignore_errors=True)

    # === CONFIGURATION ===
    def configure(
        self,
        maples_dr_path: Optional[str | Path] = UNSET,
        messidor_path: Optional[str | Path] = UNSET,
        *,
        cache: Optional[str | Path] = UNSET,
        resize: Optional[int] = None,
        image_format: Optional[ImageFormat] = None,
        preprocessing: Optional[Preprocessing] = None,
        disable_check: bool = False,
    ):
        """Configure the default behavior of the MAPLES-DR dataset.

        Any parameters left to None (or 'UNSET' for the first two paths) will leave the current configuration unaffected.

        Parameters
        ----------
        maples_dr_path : Optional[str], optional
            Path to the MAPLES-DR additional data. Must point to the directory or to the zip file.

            If None (by default), then the dataset is downloaded from figshare.

        messidor_path : Optional[str], optional
            Path to the MESSIDOR dataset.

            Must point to a directory containing the "Base11", "Base12", ... subdirectories or zip files.

        cache : Optional[str], optional
            Path to the cache directory. The cache is used to store the downloaded dataset and the generated images.

            - If ``cache`` is a ``str`` or a ``Path``, then the cache is stored in the given directory.
            - If ``False`` (by default), then the cache is disabled.
            - If ``True``, then the cache is stored in the default cache directory.

        resize : Optional[int], optional
            Set the size of the images (fundus and biomarkers) generated by `maples_dr`.

            - If ``resize`` is an int, crop the image to a square ROI and resize it to the shape ``(resize, resize)``;
            - If ``True``, keep the original MAPLES-DR resolution of 1500x1500 px;
            - If ``False``, use the original MESSIDOR resolution if MESSIDOR path is configured, otherwise fallback to MAPLES-DR original resolution.

        image_format : Optional[ImageFormat], optional
            Python format of the generated images. Must be either "PIL", "rgb" or "bgr".

            If "rgb" or "bgr" is selected, images will be formatted as numpy array of shape: (height, width, channel).

            By default, "PIL" is used.

        preprocessing : Optional[Preprocessing], optional
            Preprocessing algorithm applied on the fundus images.

            By default, no preprocessing is applied.

        disable_check : bool, optional
            If True, disable the integrity check of the dataset.

        """
        # === Update the dataset configuration ===
        self._datasets_config.update(
            dict(
                resize=resize,
                image_format=image_format,
                preprocessing=preprocessing,
            )
        )

        # === Set Maples-DR cache ===
        if cache is not UNSET and self.cfg._cache != cache:
            cache = self._change_cache_path(cache)
        else:
            cache = self.cfg.cache_path

        # === Prepare Maples-DR ===
        # If configure is called the first time with no path, download the dataset or use the cache.
        if maples_dr_path is UNSET and self._maples_dr_path is None:
            if cache is None:
                maples_dr_path = DOWNLOAD
            else:
                maples_dr_path = cache / "AdditionalData"

        if maples_dr_path is not UNSET and self._maples_dr_path != maples_dr_path:
            # Set the path and download the dataset if needed.
            maples_dr_path = self._change_maples_dr_path(maples_dr_path)

            # Load the dataset infos.
            (
                self.dataset_record,
                self._messidor_ROIs,
            ) = self.load_dataset_record_and_rois(maples_dr_path)
            self._annotations_infos = self.load_biomarkers_annotation_infos(
                maples_dr_path / "biomarkers_annotation_infos.xls"
            )

            # Check the integrity of the dataset.
            if not disable_check:
                self.check_maples_dr_integrity(
                    path=maples_dr_path,
                    biomarkers=list(self.dataset_record["biomarkers"]),
                    images_names=self.image_names(DatasetSubset.ALL_WITH_DUPLICATES),
                )

            self._diagnosis = self.load_maples_dr_diagnosis(
                maples_dr_path / "diagnosis_infos.xls"
            )

        # === Prepare MESSIDOR ===
        if messidor_path is UNSET:
            if (
                self._messidor_paths is None
                and cache is not None
                and (cache / "MESSIDOR").is_dir()
            ):
                messidor_path = cache / "MESSIDOR"

        if messidor_path is not UNSET:
            if messidor_path is None:
                self._messidor_paths = None
            else:
                self._messidor_paths = self.discover_messidor_images(
                    self.image_names(subset=DatasetSubset.ALL_WITH_DUPLICATES),
                    messidor_path,
                )

    def clear_cache(self):
        """
        Clear the cache.
        """
        if self.cfg.cache_path is not None:
            rmtree(self.cfg.cache_path, ignore_errors=True)

    @property
    def cfg(self) -> DatasetConfig:
        """
        Return the default configuration of the loaded dataset.
        """
        return self._datasets_config

    def is_configured(self) -> bool:
        """
        Check if the dataset is initialized.
        """
        return self.dataset_record is not None

    def ensure_configured(self):
        """
        Ensure the dataset is initialized.
        """
        if not self.is_initialized():
            self.configure()

    # --- Cache path configuration ---
    def _change_cache_path(self, path: Optional[str | Path | bool] = None) -> Path:
        if path is None or path is False:
            self.cfg._cache = False
            return None

        if path is True:
            path = xdg_data_home() / "maples_dr"

        # Ensure the path exists and is a directory.
        path = Path(path)
        if path.exists() and not path.is_dir():
            raise InvalidConfigError(f"Invalid cache path: {path} is not a directory.")
        path.mkdir(parents=True, exist_ok=True)

        previous_cache = self.cfg.cache_path
        if previous_cache is not None and path.absolute() == previous_cache.absolute():
            return path

        # If a cache already existed, copy the MAPLES-DR original dataset to the new cache path.
        if previous_cache is not None:
            if (previous_cache / "AdditionalData").is_dir():
                shutil.copytree(
                    previous_cache / "AdditionalData", path / "AdditionalData"
                )
            if (previous_cache / "MESSIDOR").is_dir():
                shutil.copytree(previous_cache / "MESSIDOR", path / "MESSIDOR")

        self.cfg._cache = path
        return path

    # --- Maples-DR path configuration ---
    @property
    def maples_dr_folder(self) -> Path:
        """
        Return the path to the MAPLES-DR dataset folder.
        """
        if self._maples_dr_path is None:
            raise NotConfiguredError()
        return self._maples_dr_path

    def _change_maples_dr_path(self, path: Optional[str | Path] = None) -> Path:
        """
        Return the path to the MAPLES-DR dataset folder.
        """
        # If Maples-DR was previously downloaded, delete the temporary folder.
        if self._is_maples_dr_folder_temporary and self._maples_dr_path is not None:
            rmtree(self._maples_dr_path, ignore_errors=True)
            self._is_maples_dr_folder_temporary = False

        self._maples_dr_path = None

        if path is None or path is DOWNLOAD:
            # If no path is given, prepare the dataset download to...
            path = mkdtemp()
            self._is_maples_dr_folder_temporary = True

        path = Path(path)

        if path.is_dir():
            # === If the path is a directory ===
            # Ensure it exists ...
            if not path.exists():
                path.mkdir(parents=True)
            # and check if the folder contains the dataset record.
            if not (path / "dataset_record.yaml").exists():
                # If not, download the dataset.
                zip_path = path / "maples_dr.zip"
                download(MAPLES_DR_ADDITIONAL_URL, zip_path, "MAPLES-DR labels")
                with ZipFile(path / "maples_dr.zip", "r") as zip_file:
                    zip_file.extractall(path)
                (path / "maples_dr.zip").unlink()

        elif path.name.endswith(".zip"):
            # === If the path is a zip file, unzip it to a temporary folder ===
            zip_path = path
            if not zip_path.exists():
                raise InvalidConfigError(
                    f"Invalid Maples DR archive: {zip_path} doesn't exist."
                )

            # Create a temporary folder.
            if self.cfg.cache_path is not None:
                path = self.cfg.cache_path / "AdditionalData"
            else:
                path = Path(mkdtemp())
            # Unzip the dataset to the temporary folder.
            try:
                with ZipFile(zip_path, "r") as zip_file:
                    zip_file.extractall(path)
            except Exception as e:
                raise InvalidConfigError(
                    f"Invalid Maples DR archive: {zip_path}:"
                    "\n\t the provided archive is impossible to unzip."
                ) from e
            else:
                # Test if the zip file contains maples_dr folder.
                if not (path / "dataset_record.yaml").exists():
                    raise InvalidConfigError(
                        f"Invalid Maples DR archive:  {path}:"
                        '\n\t the provided archive doesn\'t contains the file "dataset_record.yaml".'
                    )
                if not (path / "MESSIDOR-ROIs.csv").exists():
                    raise InvalidConfigError(
                        f"Invalid Maples DR archive:  {path}:"
                        '\n\t the provided archive doesn\'t contains the file "MESSIDOR-ROIs.csv".'
                    )
                if not (path / "diagnosis_infos.xls").exists():
                    raise InvalidConfigError(
                        f"Invalid Maples DR archive:  {path}:"
                        '\n\t the provided archive doesn\'t contains the file "diagnosis_infos.xls".'
                    )

                if not (path / "biomarkers_annotation_infos.xls").exists():
                    raise InvalidConfigError(
                        f"Invalid Maples DR archive:  {path}:"
                        '\n\t the provided archive doesn\'t contains the file "biomarkers_annotation_infos.xls".'
                    )
            if self.cfg.cache_path is not None:
                self._is_maples_dr_folder_temporary = True

        self._maples_dr_path = path
        return path

    @staticmethod
    def load_dataset_record_and_rois(path: str | Path) -> Tuple[Dict, Dict]:
        """
        Load the MAPLES-DR dataset record and the rois in MESSIDOR images.

        Parameters
        ----------
        path : str
            Path to the MAPLES-DR dataset folder.
        """

        with open(path / "dataset_record.yaml", "r") as infos_file:
            record = yaml.safe_load(infos_file)

        with open(path / "MESSIDOR-ROIs.csv", "r") as rois_file:
            rois = pd.read_csv(rois_file).set_index("name")

        return record, rois

    @staticmethod
    def load_biomarkers_annotation_infos(path: str | Path) -> pd.DataFrame:
        """
        Load the MAPLES-DR biomarkers annotation infos file.

        Parameters
        ----------
        path : str
            Path to the MAPLES-DR biomarkers annotation infos file.
        """

        annotation_infos_data = pd.read_excel(path, index_col=0, sheet_name=None)

        biomarker_infos = ["Retinologist", "Comment", "Time", "Annotation #"]
        biomarker_tasks = {
            "bright": BiomarkersAnnotationTasks.BRIGHT,
            "red": BiomarkersAnnotationTasks.RED,
            "disk-macula": BiomarkersAnnotationTasks.DISC_MACULA,
            "vessels": BiomarkersAnnotationTasks.VESSELS,
        }

        annotation_infos = None
        for sheet_name, task in biomarker_tasks.items():
            infos = annotation_infos_data[sheet_name][biomarker_infos].copy()
            infos["Time"] = pd.to_timedelta(infos["Time"])
            infos = infos.rename(
                columns={
                    "Retinologist": task + "_" + BiomarkersAnnotationInfos.RETINOLOGIST,
                    "Comment": task + "_" + BiomarkersAnnotationInfos.COMMENT,
                    "Time": task + "_" + BiomarkersAnnotationInfos.ANNOTATION_TIME,
                    "Annotation #": task
                    + "_"
                    + BiomarkersAnnotationInfos.ANNOTATION_ID,
                },
            )

            if annotation_infos is None:
                annotation_infos = infos
            else:
                annotation_infos = annotation_infos.join(infos)

        return annotation_infos

    def check_maples_dr_integrity(
        self, path: Path, biomarkers: list[str], images_names: list[str]
    ):
        """
        Check if the MAPLES-DR dataset contains all segmentation maps.
        """
        missing_images = 0
        for biomarker in biomarkers:
            for img in images_names:
                if (
                    self.is_biomarker_segmented(biomarker, img)
                    and not (path / "annotations" / biomarker / (img + ".png")).exists()
                ):
                    missing_images += 1
        if missing_images > 0:
            raise InvalidConfigError(
                f"The provided folder to the Maples-DR dataset is incomplete: "
                f"{missing_images} segmentation maps are missing."
            )

    @staticmethod
    def load_maples_dr_diagnosis(path: str | Path) -> pd.DataFrame:
        """
        Load the MAPLES-DR diagnostic file.

        Parameters
        ----------
        path : str
            Path to the MAPLES-DR diagnostic file.
        """
        dr_diagnosis = pd.read_excel(path, sheet_name="DR", index_col=0).rename(
            columns={"Consensus": "dr"} | {f"Retinologist{r}": f"dr_{r}" for r in "ABC"}
        )
        me_diagnosis = pd.read_excel(path, sheet_name="ME", index_col=0).rename(
            columns={"Consensus": "me"} | {f"Retinologist{r}": f"me_{r}" for r in "ABC"}
        )
        dr_me_comments = (
            pd.read_excel(path, sheet_name="Comment", index_col=0)
            .rename(columns={"Retinologist" + r: f"dr_{r}_comment" for r in "ABC"})
            .astype(str)
            .replace("nan", "")
        )
        return dr_diagnosis.join(me_diagnosis).join(dr_me_comments)

    # --- MESSIDOR path configuration ---
    @staticmethod
    def discover_messidor_images(
        images: list[str], path: Optional[str | Path] = None
    ) -> Dict[str, Path]:
        """
        Discover the MESSIDOR images corresponding to the given MAPLES-DR images.

        :param images: List of MAPLES-DR images names. The image name should not contain the extension.
        :param path: Path to the MESSIDOR dataset.
        """
        path = Path(path)
        messidor_paths = {}

        if not path.is_dir():
            raise InvalidConfigError(f"Invalid MESSIDOR path: {path} is not a folder.")

        # Scan the MESSIDOR subfolders and list images.

        # Scan MESSIDOR subfolders to find each MAPLES-DR images.
        missing_images = set(images)
        for img_path in chain.from_iterable(
            path.glob(f"**/*.{ext}") for ext in ("tif", "jpg", "png")
        ):
            try:
                missing_images.remove(img_path.stem)
            except KeyError:
                pass
            else:
                messidor_paths[img_path.stem] = img_path.absolute()
                if len(missing_images) == 0:
                    return messidor_paths

        # If some images are missing, try to unzip the MESSIDOR subfolders.
        unzip_folder = path / "maples_dr"
        unzip_folder.mkdir(parents=True, exist_ok=True)

        total_missing = len(missing_images)
        with RichProgress.iteration(
            "Retrieving MESSIDOR fundus from zip files...",
            total=total_missing,
            done_message=f"Extracted {total_missing} images in {'{t}'} second.",
        ) as progress:
            for zip_path in path.glob("**/*.zip"):
                with ZipFile(zip_path, "r") as zip_file:
                    # If the zip file contains a MESSIDOR subfolder, unzip it.
                    for zip_info in zip_file.infolist():
                        if not zip_info.is_dir() and zip_info.filename.endswith(".tif"):
                            img_filename = Path(zip_info.filename).name
                            img_stem = Path(zip_info.filename).stem
                            try:
                                missing_images.remove(img_stem)
                            except KeyError:
                                continue

                            zip_info.filename = img_filename
                            zip_file.extract(zip_info, unzip_folder)
                            messidor_paths[img_stem] = (
                                unzip_folder / img_filename
                            ).absolute()

                            progress.update(1)
                            if len(missing_images) == 0:
                                return messidor_paths

        # If some images are still missing, raise an error.
        raise InvalidConfigError(
            f"The provided folder to the MESSIDOR dataset is incomplete: "
            f"{len(missing_images)} images included in MAPLES-DR are missing."
        )

    # === DATASETS FACTORIES ===
    def load_dataset(
        self, subset: DatasetSubset | str | list[str] = DatasetSubset.ALL
    ) -> Dataset:
        """
        Return the MAPLES-DR dataset.

        :param subset: Subset of the dataset to return. If None, return the whole dataset.
                       Must be either None, "train" or "test" or a list of valid image name.
        """
        if not self.is_configured():
            self.configure()
        if isinstance(subset, list):
            # Check that all images are valid.
            valid_images = set(self.image_names(DatasetSubset.ALL_WITH_DUPLICATES))
            for img in subset:
                if img not in valid_images:
                    raise ValueError(
                        f"Invalid image name: {img} is not a valid image name."
                    )
            names = subset
        else:
            names = self.image_names(subset)

        paths = {}
        if self._messidor_paths is not None:
            paths[FundusField.FUNDUS.value] = [
                self._messidor_paths[name] for name in names
            ]

        biomarkers_folder = {
            BiomarkerField.BRIGHT_UNCERTAINS.value: "BrightUncertains",
            BiomarkerField.COTTON_WOOL_SPOTS.value: "CottonWoolSpots",
            BiomarkerField.DRUSENS.value: "Drusens",
            BiomarkerField.EXUDATES.value: "Exudates",
            BiomarkerField.HEMORRHAGES.value: "Hemorrhages",
            BiomarkerField.MACULA.value: "Macula",
            BiomarkerField.MICROANEURYSMS.value: "Microaneurysms",
            BiomarkerField.NEOVASCULARIZATION.value: "Neovascularization",
            BiomarkerField.OPTIC_CUP.value: "OpticCup",
            BiomarkerField.OPTIC_DISC.value: "OpticDisc",
            BiomarkerField.RED_UNCERTAINS.value: "RedUncertains",
            BiomarkerField.VESSELS.value: "Vessels",
        }
        for bio, bio_folder in biomarkers_folder.items():
            folder = self.maples_dr_folder / "annotations" / bio_folder
            paths[bio] = [
                folder / (n + ".png") if self.is_biomarker_segmented(bio, n) else None
                for n in names
            ]

        preannotations_folder = {
            BiomarkerField.EXUDATES.value: "Exudates",
            BiomarkerField.HEMORRHAGES.value: "Hemorrhages",
            BiomarkerField.MICROANEURYSMS.value: "Microaneurysms",
            BiomarkerField.VESSELS.value: "Vessels",
        }
        for bio, bio_folder in preannotations_folder.items():
            paths[bio + "_pre"] = [
                self.maples_dr_folder / "preannotations" / bio_folder / (name + ".png")
                for name in names
            ]

        data = pd.DataFrame(paths, index=names)

        # Add the diagnosis and annotations infos.
        data = data.join(self._diagnosis).join(self._annotations_infos)

        return Dataset(data, self.cfg, self._messidor_ROIs)

    # === UTILITIES ===
    def image_names(
        self,
        subset: DatasetSubset | str = DatasetSubset.ALL,
        extension: bool | str = False,
    ) -> list[str]:
        """
        Return the list of images names of the given subset.

        Parameters
        ----------
        subset
            Subset to return the images names from. If None, return all images names.
            Must be either None, "train", "test" or "duplicates".

        extension
            Control whether the images names should include the extension or not.
            - If False (default), return the images names without the extension.
            - If True, return the images names with a png extension.
            - If a string, return the images names with the given extension.
        """
        if not self.is_configured():
            raise NotConfiguredError()
        subset = DatasetSubset(subset)

        names = []
        if subset in (
            DatasetSubset.TRAIN,
            DatasetSubset.ALL,
            DatasetSubset.ALL_WITH_DUPLICATES,
        ):
            names += self.dataset_record["train"]
        if subset in (
            DatasetSubset.TEST,
            DatasetSubset.ALL,
            DatasetSubset.ALL_WITH_DUPLICATES,
        ):
            names += self.dataset_record["test"]
        if subset in (DatasetSubset.DUPLICATES, DatasetSubset.ALL_WITH_DUPLICATES):
            names += list(self.dataset_record["duplicates"].values())
        if extension:
            if isinstance(extension, bool):
                extension = "png"
            names = [name + "." + extension for name in names]
        return names

    def is_biomarker_segmented(
        self, biomarker: BiomarkerField | str, name: str
    ) -> bool:
        """
        Check if the given biomarker is segmented in the MAPLES-DR dataset.

        .. note::

            The macula segmentation is missing for one image centered on the optic disc.

            The optic cup boundaries are too fuzzy to be segmented on six images.


        Parameters
        ----------
        biomarker:
            The biomarker to check.
        name:
            The image name.

        Returns
        -------
        bool
            True if the biomarker is segmented, False otherwise.
        """
        if not self.is_configured():
            raise NotConfiguredError()
        biomarker = BiomarkerField.parse(biomarker)
        if (
            biomarker is BiomarkerField.MACULA
            and name in self.dataset_record["no_macula"]
        ):
            return False
        if (
            biomarker is BiomarkerField.OPTIC_CUP
            and name in self.dataset_record["no_cup"]
        ):
            return False
        return True


GLOBAL_LOADER = DatasetLoader()


def download(url: str, path: str | Path, description: Optional[str] = None):
    """
    Download the file at the given url to the given path.
    """
    response = urlopen(url)
    with RichProgress.download(
        description, byte_size=int(response.info()["Content-length"])
    ) as progress:
        with open(path, "wb") as dest_file:
            for data in iter(partial(response.read, 32768), b""):
                dest_file.write(data)
                progress.update(len(data))



from maples_dr.config import DatasetConfig


def test_config():
    cfg1 = DatasetConfig(resize=True)
    assert cfg1.resize is True

    cfg2 = DatasetConfig(resize=512)
    assert cfg2.resize == 512

    cfg1.update(dict(resize=200))
    assert cfg1.resize == 200



from __future__ import annotations

__all__ = ["RichProgress"]

import re
from enum import Enum
from functools import reduce
from pathlib import Path
from time import time
from typing import Iterable, List, Mapping, NamedTuple, Optional, Tuple, TypeGuard, TypeVar, overload

import numpy as np
import numpy.typing as npt
from rich.progress import (
    BarColumn,
    DownloadColumn,
    MofNCompleteColumn,
    Progress,
    TextColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)


class RichProgress:
    def __init__(self, name, total, done_message=None, columns=()) -> None:
        self.progress = Progress(*columns, transient=False)
        self.name = name
        self.total = total
        self.done_message = done_message
        self.task = None
        self.t0 = None

    def __enter__(self):
        self.task = self.progress.add_task(self.name, total=self.total)
        self.t0 = time()
        self.progress.start()
        return self

    def update(self, value=1, message=None):
        self.progress.update(self.task, advance=value, description=message)

    def __exit__(self, exc_type, exc_value, traceback):
        elapsed = time() - self.t0
        if exc_type is None:
            self.progress.update(self.task, visible=False)
            self.progress.remove_task(self.task)
            if self.done_message is not None:
                self.progress.console.print(self.done_message.replace("{t}", f"{elapsed:.1f}"))
        self.progress.stop()

    @staticmethod
    def iteration(name, total, done_message=None):
        return RichProgress(
            name,
            total,
            done_message,
            columns=(
                TextColumn("{task.description}"),
                BarColumn(),
                MofNCompleteColumn(),
                TimeRemainingColumn(),
            ),
        )

    @staticmethod
    def download(item_name, byte_size):
        return RichProgress(
            f"Downloading {item_name}...",
            byte_size,
            item_name + " downloaded in {t} seconds",
            columns=(
                TextColumn("{task.description}"),
                DownloadColumn(),
                TransferSpeedColumn(),
                TimeRemainingColumn(),
            ),
        )


EnumT = TypeVar("EnumT", bound=Enum)


def case_less_parse_str_enum(
    enum_type: type, value: any, ignored_characters: Optional[str] = " -_", alias: Optional[Mapping[str, EnumT]] = None
) -> EnumT:
    try:
        return enum_type(value)
    except ValueError:
        pass
    if not isinstance(value, str):
        raise ValueError(f"Invalid enum identifier: must be a string or a {type(enum_type)}, not {type(value)}.")
    if ignored_characters:
        ignored_characters = (
            ignored_characters.replace("-", "\-").replace("]", "\]").replace("^", "\^").replace("\\", "\\\\")
        )

        value = re.sub(f"[{ignored_characters}]", "", value)

    for enum in enum_type:
        name = enum.value.lower()
        if ignored_characters:
            name = name.replace("_", "")
        if name == value.lower():
            return enum
    if alias is not None:
        for name, enum in alias.items():
            if name == value.lower():
                return enum
    raise ValueError(
        f"Invalid {type(enum_type)}: {value}.\n Valid values are: {', '.join(_.value for _ in enum_type)}."
    )


class Rect(NamedTuple):
    h: float
    w: float
    y: float = 0
    x: float = 0

    @property
    def center(self) -> Point:
        return Point(self.y + self.h // 2, self.x + self.w // 2)

    @property
    def top_left(self) -> Point:
        return Point(self.y, self.x)

    @property
    def bottom_right(self) -> Point:
        return Point(self.y + self.h, self.x + self.w)

    @property
    def top(self) -> float:
        return self.y

    @property
    def bottom(self) -> float:
        return self.y + self.h

    @property
    def left(self) -> float:
        return self.x

    @property
    def right(self) -> float:
        return self.x + self.w

    @property
    def shape(self) -> Point:
        return Point(y=self.h, x=self.w)

    @property
    def area(self) -> float:
        return self.h * self.w

    def to_int(self):
        return Rect(*(int(round(_)) for _ in (self.h, self.w, self.y, self.x)))

    @classmethod
    def from_tuple(
        cls,
        rect: float
        | int
        | Tuple[float | int]
        | Tuple[float | int, float | int]
        | Tuple[float | int, float | int, float | int, float | int],
    ):
        if isinstance(rect, (float, int)):
            rect = (rect, rect)
        elif isinstance(rect, tuple) and len(rect) in (2, 4) and all(isinstance(_, (float, int)) for _ in rect):
            pass
        else:
            raise TypeError("Rect can only be created from a float or a tuple of 2 or 4 floats")
        return cls(*rect)

    @classmethod
    def from_size(cls, shape: Tuple[float | int, float | int]):
        return cls(shape[0], shape[1])

    @overload
    @classmethod
    def from_points(cls, bottom_right: Tuple[float | int, float | int]) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(cls, bottom: float | int, right: float | int) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(
        cls,
        top_left: Tuple[float | int, float | int],
        bottom_right: Tuple[float | int, float | int],
        *,
        ensure_positive: bool,
    ) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(
        cls, top: float | int, left: float | int, bottom: float | int, right: float | int, *, ensure_positive: bool
    ) -> Rect:
        ...

    @overload
    @classmethod
    def from_points(
        cls, top_left_bottom_right: Tuple[float | int, float | int, float | int, float | int], *, ensure_positive: bool
    ) -> Rect:
        ...

    @classmethod
    def from_points(
        cls,
        *p: float | int | Tuple[float | int, float | int] | Tuple[float | int, float | int, float | int, float | int],
        ensure_positive: bool = False,
    ) -> Rect:
        if (
            len(p) == 1
            and isinstance(p[0], tuple)
            and len(p[0]) == 2
            and all(isinstance(_, (float, int)) for _ in p[0])
        ):
            # case ((bottom, right), )
            p2 = Point(p[0], p[1])
            p1 = Point.origin()
        elif len(p) == 2 and all(isinstance(_, (float, int)) for _ in p):
            # case (bottom, right)
            p2 = Point(p[0], p[1])
            p1 = Point.origin()
        elif (
            len(p) == 2
            and all(isinstance(_, tuple) for _ in p)
            and all(isinstance(_, (float, int)) for _ in p[0] + p[1])
        ):
            # case ((top, left), (bottom, right))
            p1 = Point(p[0])
            p2 = Point(p[1])
        elif len(p) == 4 and all(isinstance(_, (float, int)) for _ in p):
            # case (top, left, bottom, right)
            p1 = Point(p[0], p[1])
            p2 = Point(p[2], p[3])
        elif (
            len(p) == 1
            and isinstance(p[0], tuple)
            and len(p[0]) == 4
            and all(isinstance(_, (float, int)) for _ in p[0])
        ):
            # case ((top, left, bottom, right), )
            p1 = Point(p[0])
            p2 = Point(p[1])
        else:
            raise TypeError("Rect can only be created from 2 or 4 floats or from 2 tuples of 2 floats")

        if not ensure_positive:
            return cls(abs(p2.y - p1.y), abs(p2.x - p1.x), min(p1.y, p2.y), min(p1.x, p2.x))
        else:
            rect = cls(p2.y - p1.y, p2.x - p1.x, p1.y, p1.x)
            return Rect.empty() if rect.h < 0 or rect.w < 0 else rect

    @classmethod
    def from_center(cls, center: Tuple[float, float], shape: float | Tuple[float, float]) -> Rect:
        if isinstance(shape, (float, int)):
            shape = (shape, shape)
        return cls(shape[0], shape[1], center[0] - shape[0] // 2, center[1] - shape[1] // 2)

    @classmethod
    def empty(cls) -> Rect:
        return cls(0, 0, 0, 0)

    def is_self_empty(self) -> bool:
        return self.w == 0 or self.h == 0

    @classmethod
    def is_empty(cls, rect: Rect | None) -> bool:
        if rect is None:
            return True
        if isinstance(rect, tuple) and len(rect) == 4:
            rect = Rect(*rect)
        return isinstance(rect, tuple) and (rect.w == 0 or rect.h == 0)

    @classmethod
    def is_rect(cls, r) -> TypeGuard[Rect]:
        return isinstance(r, Rect) or (isinstance(r, tuple) and len(r) == 4)

    def __repr__(self):
        return "Rect(y={}, x={}, h={}, w={})".format(self.y, self.x, self.h, self.w)

    def __or__(self, other) -> Rect:
        if isinstance(other, Rect):
            if self.is_self_empty():
                return other
            if other.is_self_empty():
                return self
            return Rect.from_points(
                (min(self.top, other.top), min(self.left, other.left)),
                (max(self.bottom, other.bottom), max(self.right, other.right)),
            )
        else:
            raise TypeError("Rect can only be combined only with another Rect")

    def __and__(self, other) -> Rect:
        if isinstance(other, Rect):
            return Rect.from_points(
                (max(self.top, other.top), max(self.left, other.left)),
                (min(self.bottom, other.bottom), min(self.right, other.right)),
                ensure_positive=True,
            )
        else:
            raise TypeError("Rect can only be combined only with another Rect")

    def __bool__(self) -> bool:
        return not self.is_self_empty()

    def __add__(self, other: Point | float) -> Rect:
        if isinstance(other, float):
            other = Point(other, other)
        if isinstance(other, Point):
            return self.translate(other.y, other.x)
        raise TypeError("Rect can only be translated by a Point or a float")

    def __sub__(self, other: Point | float) -> Rect:
        if isinstance(other, float):
            other = Point(other, other)
        if isinstance(other, Point):
            return self.translate(-other.y, -other.x)
        raise TypeError("Rect can only be translated by a Point or a float")

    def __mul__(self, other: float) -> Rect:
        return self.scale(other)

    def __truediv__(self, other: float) -> Rect:
        return self.scale(1 / other)

    def __contains__(self, other: Point | Rect) -> bool:
        if isinstance(other, Point):
            return self.y <= other.y <= self.y + self.h and self.x <= other.x <= self.x + self.w
        elif isinstance(other, Rect):
            return not Rect.is_empty(self & other)
        else:
            raise TypeError("Rect can only be compared with a Point or a Rect")

    def translate(self, y: float, x: float) -> Rect:
        return Rect(self.h, self.w, self.y + y, self.x + x)

    def scale(self, fy: float, fx: float | None = None) -> Rect:
        if fx is None:
            fx = fy
        return Rect(self.h * fy, self.w * fx, self.y * fy, self.x * fx)

    def clip(self, rect: float | Tuple[float, float] | Tuple[float, float, float, float]) -> Rect:
        rect = Rect.from_tuple(rect)
        return Rect.from_points(
            (max(self.top, rect.top), max(self.left, rect.left)),
            (min(self.bottom, rect.bottom), min(self.right, rect.right)),
            ensure_positive=True,
        )

    @overload
    def pad(self, pad: float | Tuple[float, float]) -> Rect:
        ...

    @overload
    def pad(self, vertical: float, horizontal: float) -> Rect:
        ...

    @overload
    def pad(self, top: float, right: float, bottom: float, left: float) -> Rect:
        ...

    def pad(self, *pad: float | Tuple[float, float]) -> Rect:
        if len(pad) == 1 and isinstance(pad[0], (int, float)):
            pad = (pad[0],) * 4
        elif (
            len(pad) == 1
            and isinstance(pad[0], tuple)
            and len(pad[0]) == 2
            and all(isinstance(_, (int, float)) for _ in pad[0])
        ):
            # case ((vertical, horizontal), )
            pad = (pad[0][0], pad[0][1]) * 2
        elif len(pad) == 2 and all(isinstance(_, (int, float)) for _ in pad):
            # case (vertical, horizontal)
            pad = pad * 2
        elif len(pad) == 4 and all(isinstance(_, (int, float)) for _ in pad):
            # case (top, right, bottom, left)
            pass
        else:
            raise TypeError("Rect.pad() only accept 1, 2 or 4 floats as arguments")

        return Rect(self.h + pad[0] + pad[2], self.w + pad[1] + pad[3], self.y - pad[0], self.x - pad[3])

    def box(self) -> tuple[int, int, int, int]:
        return self.left, self.top, self.right, self.bottom

    def slice(self) -> tuple[slice, slice]:
        r = self.to_int()
        return slice(r.y, r.y + r.h), slice(r.x, r.x + r.w)

    @staticmethod
    def union(*rects: Tuple[Iterable[Rect] | Rect, ...]) -> Rect:
        rects = sum(((r,) if isinstance(r, Rect) else tuple(r) for r in rects), ())
        return reduce(lambda a, b: a | b, rects)

    @staticmethod
    def intersection(*rects: Tuple[Iterable[Rect] | Rect, ...]) -> Rect:
        rects = sum(((r,) if isinstance(r, Rect) else tuple(r) for r in rects), ())
        return reduce(lambda a, b: a & b, rects)


class Point(NamedTuple):
    y: float
    x: float

    def xy(self) -> tuple[float, float]:
        return self.x, self.y

    def to_int(self) -> Point:
        return Point(int(round(self.y)), int(round(self.x)))

    def clip(self, rect: float | Tuple[float, float] | Tuple[float, float, float, float]) -> Point:
        rect = Rect.from_tuple(rect)
        return Point(
            min(max(self.y, rect.top), rect.bottom),
            min(max(self.x, rect.left), rect.right),
        )

    def __add__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y + other, self.x + other)
        y, x = other
        return Point(self.y + y, self.x + x)

    def __radd__(self, other: Tuple[float, float] | float):
        return self + other

    def __sub__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y - other, self.x - other)
        y, x = other
        return Point(self.y - y, self.x - x)

    def __rsub__(self, other: Tuple[float, float] | float):
        return -(self - other)

    def __mul__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y * other, self.x * other)
        y, x = other
        return Point(self.y * y, self.x * x)

    def __rmul__(self, other: Tuple[float, float] | float):
        return self * other

    def __truediv__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(self.y / other, self.x / other)
        y, x = other
        return Point(self.y / y, self.x / x)

    def __rtruediv__(self, other: Tuple[float, float] | float):
        if isinstance(other, (float, int)):
            return Point(other / self.y, other / self.x)
        y, x = other
        return Point(y / self.y, x / self.x)

    def __neg__(self):
        return Point(-self.y, -self.x)

    @classmethod
    def origin(cls):
        return cls(0, 0)

    @classmethod
    def from_tuple(cls, point: float | int | Tuple[float | int] | Tuple[float | int, float | int]):
        if isinstance(point, (float, int)):
            return cls(point, point)
        if len(point) == 1:
            if isinstance(point[0], (float, int)):
                return cls(point[0], point[0])
            if isinstance(point[0], tuple):
                return cls(*point[0])
        if len(point) == 2:
            return cls(*point)
        raise TypeError("Point can only be created from a float or a tuple of 2 floats")

    @overload
    def distance(self, other: Point) -> float:
        ...

    @overload
    def distance(self, other: List[Point]) -> List[float]:
        ...

    @overload
    def distance(self, other: npt.NDArray[np.float]) -> npt.NDArray[np.float]:
        ...

    def distance(self, other: Point | Iterable[Point]) -> float | Iterable[float]:
        import numpy as np

        if isinstance(other, np.ndarray):
            return np.linalg.norm(other - self, axis=-1)
        elif isinstance(other, list):
            return [self.distance(p) for p in other]
        return ((self.y - other.y) ** 2 + (self.x - other.x) ** 2) ** 0.5


def xdg_data_home() -> str:
    import os

    if os.name == "nt":
        path = os.environ.get("APPDATA")
    else:
        path = os.environ.get("XDG_DATA_HOME")
    return Path.home() / ".cache" if path is None else Path(path).absolute()



import maples_dr
import numpy as np
from maples_dr.dataset import FundusField as FF
from pytest import fixture


@fixture
def all_set():
    maples_dr.configure(
        maples_dr_path="examples/PATH/TO/MAPLES-DR/AdditionalData.zip",
        messidor_path="examples/PATH/TO/MESSIDOR/",
    )
    return maples_dr.quick_api.GLOBAL_LOADER.load_dataset("all_with_duplicates")


def test_fundus(all_set):
    # === Check data content ===
    assert "fundus" in all_set.data

    sample = all_set[0]
    sample._cfg.image_format = "bgr"

    # Test equivalence of __getitem__ and actual read methods for fundus and mask
    assert np.all(sample.read_fundus(preprocess=False) == sample[FF.RAW_FUNDUS])
    assert np.all(sample.read_roi_mask() == sample[FF.MASK])



import maples_dr


def test_configure():
    maples_dr.configure(
        maples_dr_path="examples/PATH/TO/MAPLES-DR/AdditionalData.zip",
        messidor_path="examples/PATH/TO/MESSIDOR/",
        resize=512,
    )

    test_set = maples_dr.load_test_set()
    assert test_set._cfg.resize == 512

    train_set = maples_dr.load_train_set()
    assert test_set._cfg is train_set._cfg

    maples_dr.configure(resize=1024)
    assert test_set._cfg.resize == 1024
